#!/usr/bin/env python3
#
# Copyright 2019 The Kapitan Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging
import os
import yaml
import json
import collections

from kapitan.errors import CompileError, KapitanError
from kapitan.refs.base import Revealer
from kapitan.utils import PrettyDumper

logger = logging.getLogger(__name__)


class InputType(object):
    def __init__(self, type_name, compile_path, search_paths, ref_controller):
        self.type_name = type_name
        self.compile_path = compile_path
        self.search_paths = search_paths
        self.ref_controller = ref_controller

    def compile_obj(self, comp_obj, ext_vars, **kwargs):
        """
        run compile_input_path() for each input_path in comp_obj
        kwargss are passed into compile_input_path()
        """
        input_type = comp_obj["input_type"]
        assert input_type == self.type_name
        input_paths = comp_obj["input_paths"]

        for input_path in input_paths:
            self.compile_input_path(input_path, comp_obj, ext_vars, **kwargs)

    def compile_input_path(self, input_path, comp_obj, ext_vars, **kwargs):
        """
        compile and validate input_path in comp_obj
        kwargs are passed into compile_file()
        """
        target_name = ext_vars["target"]
        output_path = comp_obj["output_path"]
        output_type = comp_obj.get("output_type", self.default_output_type())
        file_found = False

        for path in self.search_paths:
            compile_file_sp = os.path.join(path, input_path)
            if os.path.exists(compile_file_sp):
                file_found = True
                logger.debug("Compiling %s", compile_file_sp)
                try:
                    _compile_path = os.path.join(self.compile_path, target_name, output_path)
                    self.compile_file(compile_file_sp, _compile_path, ext_vars, output=output_type,
                                      target_name=target_name, **kwargs)
                except KapitanError as e:
                    raise CompileError("{}\nCompile error: failed to compile target: {}".format(e, target_name))

        if not file_found:
            raise CompileError("Compile error: {} for target: {} not found in "
                               "search_paths: {}".format(input_path, target_name, self.search_paths))

    def make_compile_dirs(self, target_name, output_path):
        """make compile dirs, skips if dirs exist"""
        _compile_path = os.path.join(self.compile_path, target_name, output_path)
        # support writing to an already existent dir
        os.makedirs(_compile_path, exist_ok=True)

    def compile_file(self, file_path, compile_path, ext_vars, **kwargs):
        """implements compilation for file_path to compile_path with ext_vars"""
        return NotImplementedError

    def default_output_type(self):
        "returns default output_type value"
        return NotImplementedError


class CompilingFile(object):
    def __init__(self, context, fp, ref_controller, **kwargs):
        self.fp = fp
        self.ref_controller = ref_controller
        self.kwargs = kwargs
        self.revealer = Revealer(ref_controller)

    def write(self, data):
        """write data into file"""
        reveal = self.kwargs.get('reveal', False)
        target_name = self.kwargs.get('target_name', None)
        if reveal:
            self.fp.write(self.revealer.reveal_raw(data))
        else:
            self.fp.write(self.revealer.compile_raw(data, target_name=target_name))

    def write_yaml(self, obj):
        """recursively compile or reveal refs and convert obj to yaml and write to file"""
        indent = self.kwargs.get('indent', 2)
        reveal = self.kwargs.get('reveal', False)
        target_name = self.kwargs.get('target_name', None)
        if reveal:
            self.revealer.reveal_obj(obj)
        else:
            self.revealer.compile_obj(obj, target_name=target_name)

        if isinstance(obj, collections.Mapping):
            yaml.dump(obj, stream=self.fp, indent=indent, Dumper=PrettyDumper, default_flow_style=False)
        else:
            yaml.dump_all(obj, stream=self.fp, indent=indent, Dumper=PrettyDumper, default_flow_style=False)

        logger.debug("Wrote %s", self.fp.name)

    def write_json(self, obj):
        """recursively hash or reveal refs and convert obj to json and write to file"""
        indent = self.kwargs.get('indent', 2)
        reveal = self.kwargs.get('reveal', False)
        target_name = self.kwargs.get('target_name', None)
        if reveal:
            self.revealer.reveal_obj(obj)
        else:
            self.revealer.compile_obj(obj, target_name=target_name)
        json.dump(obj, self.fp, indent=indent)
        logger.debug("Wrote %s", self.fp.name)


class CompiledFile(object):
    def __init__(self, name, ref_controller, **kwargs):
        self.name = name
        self.fp = None
        self.ref_controller = ref_controller
        self.kwargs = kwargs

    def __enter__(self):
        mode = self.kwargs.get("mode", "r")
        self.fp = open(self.name, mode)
        return CompilingFile(self, self.fp, self.ref_controller, **self.kwargs)

    def __exit__(self, *args):
        self.fp.close()



#!/usr/bin/env python3
#
# Copyright 2019 The Kapitan Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"compile tests"

import unittest
import os
import sys
import io
import contextlib
from kapitan.cli import main
from kapitan.utils import directory_hash
from kapitan.cached import reset_cache
from kapitan.targets import validate_matching_target_name
from kapitan.resources import inventory_reclass
from kapitan.errors import InventoryError


class CompileKubernetesTest(unittest.TestCase):
    def setUp(self):
        os.chdir(os.getcwd() + '/examples/kubernetes/')

    def test_compile(self):
        sys.argv = ["kapitan", "compile", "-c"]
        main()
        os.remove('./compiled/.kapitan_cache')
        compiled_dir_hash = directory_hash(os.getcwd() + '/compiled')
        test_compiled_dir_hash = directory_hash(os.getcwd() + '/../../tests/test_kubernetes_compiled')
        self.assertEqual(compiled_dir_hash, test_compiled_dir_hash)

    def test_compile_not_enough_args(self):
        with self.assertRaises(SystemExit) as cm:
            # Ignoring stdout for "kapitan --help"
            with contextlib.redirect_stdout(io.StringIO()):
                sys.argv = ["kapitan"]
                main()
        self.assertEqual(cm.exception.code, 1)

    def test_compile_not_matching_targets(self):
        with self.assertLogs(logger='kapitan.targets', level='ERROR') as cm, contextlib.redirect_stdout(io.StringIO()):
            # as of now, we cannot capture stdout with contextlib.redirect_stdout
            # since we only do logger.error(e) in targets.py before exiting
            with self.assertRaises(SystemExit) as ca:
                unmatched_filename = "inventory/targets/minikube-es-fake.yml"
                correct_filename = "inventory/targets/minikube-es.yml"
                os.rename(src=correct_filename, dst=unmatched_filename)
                sys.argv = ["kapitan", "compile"]

                try:
                    main()
                finally:
                    # correct the filename again, even if assertion fails
                    if os.path.exists(unmatched_filename):
                        os.rename(src=unmatched_filename, dst=correct_filename)
        error_message_substr = "is missing the corresponding yml file"
        self.assertTrue(' '.join(cm.output).find(error_message_substr) != -1)

    def test_compile_vars_target_missing(self):
        inventory_path = "inventory"
        target_filename = "minikube-es"
        target_obj = inventory_reclass(inventory_path)['nodes'][target_filename]['parameters']['kapitan']
        # delete vars.target
        del target_obj["vars"]["target"]

        with self.assertRaises(InventoryError) as ie:
            validate_matching_target_name(target_filename, target_obj, inventory_path)

        error_message = "Target missing: target \"{}\" is missing parameters.kapitan.vars.target\n" \
                        "This parameter should be set to the target name"
        self.assertTrue(error_message.format(target_filename), ie.exception.args[0])

    def tearDown(self):
        os.chdir(os.getcwd() + '/../../')
        reset_cache()


class CompileTerraformTest(unittest.TestCase):
    def setUp(self):
        os.chdir(os.getcwd() + '/examples/terraform/')

    def test_compile(self):
        sys.argv = ["kapitan", "compile"]
        main()
        compiled_dir_hash = directory_hash(os.getcwd() + '/compiled')
        test_compiled_dir_hash = directory_hash(
            os.getcwd() + '/../../tests/test_terraform_compiled')
        self.assertEqual(compiled_dir_hash, test_compiled_dir_hash)

    def tearDown(self):
        os.chdir(os.getcwd() + '/../../')
        reset_cache()


class PlainOutputTest(unittest.TestCase):
    def setUp(self):
        os.chdir(os.getcwd() + '/examples/docker/')

    def test_compile(self):
        sys.argv = ["kapitan", "compile"]
        main()
        compiled_dir_hash = directory_hash(os.getcwd() + '/compiled')
        test_compiled_dir_hash = directory_hash(
            os.getcwd() + '/../../tests/test_docker_compiled')
        self.assertEqual(compiled_dir_hash, test_compiled_dir_hash)

    def tearDown(self):
        os.chdir(os.getcwd() + '/../../')
        reset_cache()



#!/usr/bin/env python3
#
# Copyright 2019 The Kapitan Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"command line module"

from __future__ import print_function

import argparse
import base64
import json
import logging
import os
import sys

import yaml
from kapitan import cached
from kapitan.errors import KapitanError, RefHashMismatchError
from kapitan.initialiser import initialise_skeleton
from kapitan.lint import start_lint
from kapitan.refs.base import PlainRef, RefController, Revealer
from kapitan.refs.base64 import Base64Ref
from kapitan.refs.secrets.awskms import AWSKMSSecret
from kapitan.refs.secrets.gkms import GoogleKMSSecret
from kapitan.refs.secrets.vaultkv import VaultSecret
from kapitan.refs.secrets.gpg import GPGSecret, lookup_fingerprints
from kapitan.resources import (inventory_reclass, resource_callbacks,
                               search_imports)
from kapitan.targets import compile_targets, schema_validate_compiled
from kapitan.inputs.jinja2_filters import DEFAULT_JINJA2_FILTERS_PATH
from kapitan.utils import (PrettyDumper, check_version, deep_get, fatal_error,
                           flatten_dict, from_dot_kapitan, jsonnet_file,
                           search_target_token_paths, searchvar)
from kapitan.version import DESCRIPTION, PROJECT_NAME, VERSION

logger = logging.getLogger(__name__)


def main():
    """main function for command line usage"""
    parser = argparse.ArgumentParser(prog=PROJECT_NAME,
                                     description=DESCRIPTION)
    parser.add_argument('--version', action='version', version=VERSION)
    subparser = parser.add_subparsers(help="commands")

    eval_parser = subparser.add_parser('eval', help='evaluate jsonnet file')
    eval_parser.add_argument('jsonnet_file', type=str)
    eval_parser.add_argument('--output', type=str,
                             choices=('yaml', 'json'),
                             default=from_dot_kapitan('eval', 'output', 'yaml'),
                             help='set output format, default is "yaml"')
    eval_parser.add_argument('--vars', type=str,
                             default=from_dot_kapitan('eval', 'vars', []),
                             nargs='*',
                             metavar='VAR',
                             help='set variables')
    eval_parser.add_argument('--search-paths', '-J', type=str, nargs='+',
                             default=from_dot_kapitan('eval', 'search-paths', ['.']),
                             metavar='JPATH',
                             help='set search paths, default is ["."]')

    compile_parser = subparser.add_parser('compile', help='compile targets')
    compile_parser.add_argument('--search-paths', '-J', type=str, nargs='+',
                                default=from_dot_kapitan('compile', 'search-paths', ['.', 'lib']),
                                metavar='JPATH',
                                help='set search paths, default is ["."]')
    compile_parser.add_argument('--jinja2-filters', '-J2F', type=str,
                                default=from_dot_kapitan('compile', 'jinja2-filters',
                                                         DEFAULT_JINJA2_FILTERS_PATH),
                                metavar='FPATH',
                                help='load custom jinja2 filters from any file, default is to put\
                                them inside lib/jinja2_filters.py')
    compile_parser.add_argument('--verbose', '-v', help='set verbose mode',
                                action='store_true',
                                default=from_dot_kapitan('compile', 'verbose', False))
    compile_parser.add_argument('--prune', help='prune jsonnet output',
                                action='store_true',
                                default=from_dot_kapitan('compile', 'prune', False))
    compile_parser.add_argument('--quiet', help='set quiet mode, only critical output',
                                action='store_true',
                                default=from_dot_kapitan('compile', 'quiet', False))
    compile_parser.add_argument('--output-path', type=str,
                                default=from_dot_kapitan('compile', 'output-path', '.'),
                                metavar='PATH',
                                help='set output path, default is "."')
    compile_parser.add_argument('--fetch',
                                help='fetches external dependencies', action='store_true',
                                default=from_dot_kapitan('compile', 'fetch', False))
    compile_parser.add_argument('--validate',
                                help='validate compile output against schemas as specified in inventory',
                                action='store_true', default=from_dot_kapitan('compile', 'validate', False))
    compile_parser.add_argument('--targets', '-t', help='targets to compile, default is all',
                                type=str, nargs='+',
                                default=from_dot_kapitan('compile', 'targets', []),
                                metavar='TARGET')
    compile_parser.add_argument('--parallelism', '-p', type=int,
                                default=from_dot_kapitan('compile', 'parallelism', 4),
                                metavar='INT',
                                help='Number of concurrent compile processes, default is 4')
    compile_parser.add_argument('--indent', '-i', type=int,
                                default=from_dot_kapitan('compile', 'indent', 2),
                                metavar='INT',
                                help='Indentation spaces for YAML/JSON, default is 2')
    compile_parser.add_argument('--refs-path', help='set refs path, default is "./refs"',
                                default=from_dot_kapitan('compile', 'refs-path', './refs'))
    compile_parser.add_argument('--reveal',
                                help='reveal refs (warning: this will potentially write sensitive data)',
                                action='store_true',
                                default=from_dot_kapitan('compile', 'reveal', False))
    compile_parser.add_argument('--inventory-path',
                                default=from_dot_kapitan('compile', 'inventory-path', './inventory'),
                                help='set inventory path, default is "./inventory"')
    compile_parser.add_argument('--cache', '-c',
                                help='enable compilation caching to .kapitan_cache, default is False',
                                action='store_true',
                                default=from_dot_kapitan('compile', 'cache', False))
    compile_parser.add_argument('--cache-paths', type=str, nargs='+',
                                default=from_dot_kapitan('compile', 'cache-paths', []),
                                metavar='PATH',
                                help='cache additional paths to .kapitan_cache, default is []')
    compile_parser.add_argument('--ignore-version-check',
                                help='ignore the version from .kapitan',
                                action='store_true',
                                default=from_dot_kapitan('compile', 'ignore-version-check', False))
    compile_parser.add_argument('--schemas-path',
                                default=from_dot_kapitan('validate', 'schemas-path', './schemas'),
                                help='set schema cache path, default is "./schemas"')

    inventory_parser = subparser.add_parser('inventory', help='show inventory')
    inventory_parser.add_argument('--target-name', '-t',
                                  default=from_dot_kapitan('inventory', 'target-name', ''),
                                  help='set target name, default is all targets')
    inventory_parser.add_argument('--inventory-path',
                                  default=from_dot_kapitan('inventory', 'inventory-path', './inventory'),
                                  help='set inventory path, default is "./inventory"')
    inventory_parser.add_argument('--flat', '-F', help='flatten nested inventory variables',
                                  action='store_true',
                                  default=from_dot_kapitan('inventory', 'flat', False))
    inventory_parser.add_argument('--pattern', '-p',
                                  default=from_dot_kapitan('inventory', 'pattern', ''),
                                  help='filter pattern (e.g. parameters.mysql.storage_class, or storage_class,' +
                                  ' or storage_*), default is ""')
    inventory_parser.add_argument('--verbose', '-v', help='set verbose mode',
                                  action='store_true',
                                  default=from_dot_kapitan('inventory', 'verbose', False))

    searchvar_parser = subparser.add_parser('searchvar',
                                            help='show all inventory files where var is declared')
    searchvar_parser.add_argument('searchvar', type=str, metavar='VARNAME',
                                  help='e.g. parameters.mysql.storage_class, or storage_class, or storage_*')
    searchvar_parser.add_argument('--inventory-path',
                                  default=from_dot_kapitan('searchvar', 'inventory-path', './inventory'),
                                  help='set inventory path, default is "./inventory"')
    searchvar_parser.add_argument('--verbose', '-v', help='set verbose mode',
                                  action='store_true',
                                  default=from_dot_kapitan('searchvar', 'verbose', False))
    searchvar_parser.add_argument('--pretty-print', '-p', help='Pretty print content of var',
                                  action='store_true',
                                  default=from_dot_kapitan('searchvar', 'pretty-print', False))

    subparser.add_parser('secrets', help='(DEPRECATED) please use refs')

    refs_parser = subparser.add_parser('refs', help='manage refs')
    refs_parser.add_argument('--write', '-w', help='write ref token',
                             metavar='TOKENNAME',)
    refs_parser.add_argument('--update', help='update GPG recipients for ref token',
                             metavar='TOKENNAME',)
    refs_parser.add_argument('--update-targets', action='store_true',
                             default=from_dot_kapitan('refs', 'update-targets', False),
                             help='update target secret refs')
    refs_parser.add_argument('--validate-targets', action='store_true',
                             default=from_dot_kapitan('refs', 'validate-targets', False),
                             help='validate target secret refs')
    refs_parser.add_argument('--base64', '-b64', help='base64 encode file content',
                             action='store_true',
                             default=from_dot_kapitan('refs', 'base64', False))
    refs_parser.add_argument('--reveal', '-r', help='reveal refs',
                             action='store_true',
                             default=from_dot_kapitan('refs', 'reveal', False))
    refs_parser.add_argument('--file', '-f', help='read file or directory, set "-" for stdin',
                             metavar='FILENAME')
    refs_parser.add_argument('--target-name', '-t', help='grab recipients from target name')
    refs_parser.add_argument('--inventory-path',
                             default=from_dot_kapitan('refs', 'inventory-path', './inventory'),
                             help='set inventory path, default is "./inventory"')
    refs_parser.add_argument('--recipients', '-R', help='set GPG recipients',
                             type=str, nargs='+',
                             default=from_dot_kapitan('refs', 'recipients', []),
                             metavar='RECIPIENT')
    refs_parser.add_argument('--key', '-K', help='set KMS key',
                             default=from_dot_kapitan('refs', 'key', ''),
                             metavar='KEY')
    refs_parser.add_argument('--vault-auth', help='set authentication type for vaultkv secrets',
                             default=from_dot_kapitan('refs', 'vault-auth', ''),
                             metavar='AUTH')
    refs_parser.add_argument('--refs-path', help='set refs path, default is "./refs"',
                             default=from_dot_kapitan('refs', 'refs-path', './refs'))
    refs_parser.add_argument('--verbose', '-v',
                             help='set verbose mode (warning: this will potentially show sensitive data)',
                             action='store_true',
                             default=from_dot_kapitan('refs', 'verbose', False))

    lint_parser = subparser.add_parser('lint', help='linter for inventory and refs')
    lint_parser.add_argument('--fail-on-warning',
                             default=from_dot_kapitan('lint', 'fail-on-warning', False),
                             action='store_true',
                             help='exit with failure code if warnings exist, default is False')
    lint_parser.add_argument('--skip-class-checks',
                             action='store_true',
                             help='skip checking for unused classes, default is False',
                             default=from_dot_kapitan('lint', 'skip-class-checks', False))
    lint_parser.add_argument('--skip-yamllint',
                             action='store_true',
                             help='skip running yamllint on inventory, default is False',
                             default=from_dot_kapitan('lint', 'skip-yamllint', False))
    lint_parser.add_argument('--search-secrets',
                             default=from_dot_kapitan('lint', 'search-secrets', False),
                             action='store_true',
                             help='searches for plaintext secrets in inventory, default is False')
    lint_parser.add_argument('--refs-path',
                             help='set refs path, default is "./refs"',
                             default=from_dot_kapitan('lint', 'refs-path', './refs'))
    lint_parser.add_argument('--compiled-path',
                             default=from_dot_kapitan('lint', 'compiled-path', './compiled'),
                             help='set compiled path, default is "./compiled"')
    lint_parser.add_argument('--inventory-path',
                             default=from_dot_kapitan('lint', 'inventory-path', './inventory'),
                             help='set inventory path, default is "./inventory"')

    init_parser = subparser.add_parser('init',
                                       help='initialize a directory with the recommended kapitan project skeleton.')
    init_parser.add_argument('--directory',
                             default=from_dot_kapitan('init', 'directory', '.'),
                             help='set path, in which to generate the project skeleton,'
                             'assumes directory already exists. default is "./"')

    validate_parser = subparser.add_parser('validate',
                                           help='validates the compile output against schemas as specified in inventory')
    validate_parser.add_argument('--compiled-path',
                                 default=from_dot_kapitan('compile', 'compiled-path', './compiled'),
                                 help='set compiled path, default is "./compiled')
    validate_parser.add_argument('--inventory-path',
                                 default=from_dot_kapitan('compile', 'inventory-path', './inventory'),
                                 help='set inventory path, default is "./inventory"')
    validate_parser.add_argument('--targets', '-t', help='targets to validate, default is all',
                                 type=str, nargs='+',
                                 default=from_dot_kapitan('compile', 'targets', []),
                                 metavar='TARGET'),
    validate_parser.add_argument('--schemas-path',
                                 default=from_dot_kapitan('validate', 'schemas-path', './schemas'),
                                 help='set schema cache path, default is "./schemas"')
    validate_parser.add_argument('--parallelism', '-p', type=int,
                                 default=from_dot_kapitan('validate', 'parallelism', 4),
                                 metavar='INT',
                                 help='Number of concurrent validate processes, default is 4')
    args = parser.parse_args()

    logger.debug('Running with args: %s', args)

    try:
        cmd = sys.argv[1]
    except IndexError:
        parser.print_help()
        sys.exit(1)

    # cache args where key is subcommand
    cached.args[sys.argv[1]] = args

    if hasattr(args, 'verbose') and args.verbose:
        logging.basicConfig(level=logging.DEBUG,
                            format='%(asctime)s %(name)-12s %(levelname)-8s %(message)s')
    elif hasattr(args, 'quiet') and args.quiet:
        logging.basicConfig(level=logging.CRITICAL, format="%(message)s")
    else:
        logging.basicConfig(level=logging.INFO, format="%(message)s")

    if cmd == 'eval':
        file_path = args.jsonnet_file
        search_paths = [os.path.abspath(path) for path in args.search_paths]
        ext_vars = {}
        if args.vars:
            ext_vars = dict(var.split('=') for var in args.vars)
        json_output = None

        def _search_imports(cwd, imp):
            return search_imports(cwd, imp, search_paths)

        json_output = jsonnet_file(file_path, import_callback=_search_imports,
                                   native_callbacks=resource_callbacks(search_paths),
                                   ext_vars=ext_vars)
        if args.output == 'yaml':
            json_obj = json.loads(json_output)
            yaml.safe_dump(json_obj, sys.stdout, default_flow_style=False)
        elif json_output:
            print(json_output)

    elif cmd == 'compile':
        search_paths = [os.path.abspath(path) for path in args.search_paths]

        if not args.ignore_version_check:
            check_version()

        ref_controller = RefController(args.refs_path)
        # cache controller for use in reveal_maybe jinja2 filter
        cached.ref_controller_obj = ref_controller
        cached.revealer_obj = Revealer(ref_controller)

        compile_targets(args.inventory_path, search_paths, args.output_path,
                        args.parallelism, args.targets, ref_controller,
                        prune=(args.prune), indent=args.indent, reveal=args.reveal,
                        cache=args.cache, cache_paths=args.cache_paths,
                        fetch_dependencies=args.fetch, validate=args.validate,
                        schemas_path=args.schemas_path,
                        jinja2_filters=args.jinja2_filters)

    elif cmd == 'inventory':
        if args.pattern and args.target_name == '':
            parser.error("--pattern requires --target_name")
        try:
            inv = inventory_reclass(args.inventory_path)
            if args.target_name != '':
                inv = inv['nodes'][args.target_name]
                if args.pattern != '':
                    pattern = args.pattern.split(".")
                    inv = deep_get(inv, pattern)
            if args.flat:
                inv = flatten_dict(inv)
                yaml.dump(inv, sys.stdout, width=10000, default_flow_style=False)
            else:
                yaml.dump(inv, sys.stdout, Dumper=PrettyDumper, default_flow_style=False)
        except Exception as e:
            if not isinstance(e, KapitanError):
                logger.exception("\n~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~")
            sys.exit(1)

    elif cmd == 'searchvar':
        searchvar(args.searchvar, args.inventory_path, args.pretty_print)

    elif cmd == 'lint':
        start_lint(args.fail_on_warning, args.skip_class_checks, args.skip_yamllint, args.inventory_path,
                   args.search_secrets, args.refs_path, args.compiled_path)

    elif cmd == 'init':
        initialise_skeleton(args.directory)

    elif cmd == 'secrets':
        logger.error("Secrets have been renamed to refs, please refer to: '$ kapitan refs --help'")
        sys.exit(1)

    elif cmd == 'refs':
        ref_controller = RefController(args.refs_path)

        if args.write is not None:
            ref_write(args, ref_controller)
        elif args.reveal:
            ref_reveal(args, ref_controller)
        elif args.update:
            secret_update(args, ref_controller)
        elif args.update_targets or args.validate_targets:
            secret_update_validate(args, ref_controller)

    elif cmd == 'validate':
        schema_validate_compiled(args.targets, inventory_path=args.inventory_path, compiled_path=args.compiled_path,
                                 schema_cache_path=args.schemas_path, parallel=args.parallelism)


def ref_write(args, ref_controller):
    "Write ref to ref_controller based on cli args"
    token_name = args.write
    file_name = args.file
    data = None

    if file_name is None:
        fatal_error('--file is required with --write')
    if file_name == '-':
        data = ''
        for line in sys.stdin:
            data += line
    else:
        with open(file_name) as fp:
            data = fp.read()

    if token_name.startswith("gpg:"):
        type_name, token_path = token_name.split(":")
        recipients = [dict((("name", name),)) for name in args.recipients]
        if args.target_name:
            inv = inventory_reclass(args.inventory_path)
            kap_inv_params = inv['nodes'][args.target_name]['parameters']['kapitan']
            if 'secrets' not in kap_inv_params:
                raise KapitanError("parameters.kapitan.secrets not defined in {}".format(args.target_name))

            recipients = kap_inv_params['secrets']['gpg']['recipients']
        if not recipients:
            raise KapitanError("No GPG recipients specified. Use --recipients or specify them in " +
                               "parameters.kapitan.secrets.gpg.recipients and use --target-name")
        secret_obj = GPGSecret(data, recipients, encode_base64=args.base64)
        tag = '?{{gpg:{}}}'.format(token_path)
        ref_controller[tag] = secret_obj

    elif token_name.startswith("gkms:"):
        type_name, token_path = token_name.split(":")
        key = args.key
        if args.target_name:
            inv = inventory_reclass(args.inventory_path)
            kap_inv_params = inv['nodes'][args.target_name]['parameters']['kapitan']
            if 'secrets' not in kap_inv_params:
                raise KapitanError("parameters.kapitan.secrets not defined in {}".format(args.target_name))

            key = kap_inv_params['secrets']['gkms']['key']
        if not key:
            raise KapitanError("No KMS key specified. Use --key or specify it in parameters.kapitan.secrets.gkms.key and use --target-name")
        secret_obj = GoogleKMSSecret(data, key, encode_base64=args.base64)
        tag = '?{{gkms:{}}}'.format(token_path)
        ref_controller[tag] = secret_obj

    elif token_name.startswith("awskms:"):
        type_name, token_path = token_name.split(":")
        key = args.key
        if args.target_name:
            inv = inventory_reclass(args.inventory_path)
            kap_inv_params = inv['nodes'][args.target_name]['parameters']['kapitan']
            if 'secrets' not in kap_inv_params:
                raise KapitanError("parameters.kapitan.secrets not defined in {}".format(args.target_name))

            key = kap_inv_params['secrets']['awskms']['key']
        if not key:
            raise KapitanError("No KMS key specified. Use --key or specify it in parameters.kapitan.secrets.awskms.key and use --target-name")
        secret_obj = AWSKMSSecret(data, key, encode_base64=args.base64)
        tag = '?{{awskms:{}}}'.format(token_path)
        ref_controller[tag] = secret_obj

    elif token_name.startswith("base64:"):
        type_name, token_path = token_name.split(":")
        _data = data.encode()
        encoding = 'original'
        if args.base64:
            _data = base64.b64encode(_data).decode()
            _data = _data.encode()
            encoding = 'base64'
        ref_obj = Base64Ref(_data, encoding=encoding)
        tag = '?{{base64:{}}}'.format(token_path)
        ref_controller[tag] = ref_obj

    elif token_name.startswith("vaultkv:"):
        type_name, token_path = token_name.split(":")
        _data = data.encode()
        vault_params = {}
        encoding = 'original'
        if args.target_name:
            inv = inventory_reclass(args.inventory_path)
            kap_inv_params = inv['nodes'][args.target_name]['parameters']['kapitan']
            if 'secrets' not in kap_inv_params:
                raise KapitanError("parameters.kapitan.secrets not defined in {}".format(args.target_name))

            vault_params = kap_inv_params['secrets']['vaultkv']
        if args.vault_auth:
            vault_params['auth'] = args.vault_auth
        if vault_params.get('auth') is None:
            raise KapitanError("No Authentication type parameter specified. Specify it"
                               " in parameters.kapitan.secrets.vaultkv.auth and use --target-name or use --vault-auth")

        secret_obj = VaultSecret(_data, vault_params)
        tag = '?{{vaultkv:{}}}'.format(token_path)
        ref_controller[tag] = secret_obj

    elif token_name.startswith("plain:"):
        type_name, token_path = token_name.split(":")
        _data = data.encode()
        encoding = 'original'
        if args.base64:
            _data = base64.b64encode(_data).decode()
            _data = _data.encode()
            encoding = 'base64'
        ref_obj = PlainRef(_data, encoding=encoding)
        tag = '?{{plain:{}}}'.format(token_path)
        ref_controller[tag] = ref_obj

    else:
        fatal_error("Invalid token: {name}. Try using gpg/gkms/awskms/vaultkv/base64/plain:{name}".format(name=token_name))


def secret_update(args, ref_controller):
    "Update secret gpg recipients/gkms/awskms key"
    # TODO --update *might* mean something else for other types
    token_name = args.update
    if token_name.startswith("gpg:"):
        # args.recipients is a list, convert to recipients dict
        recipients = [dict([("name", name), ]) for name in args.recipients]
        if args.target_name:
            inv = inventory_reclass(args.inventory_path)
            kap_inv_params = inv['nodes'][args.target_name]['parameters']['kapitan']
            if 'secrets' not in kap_inv_params:
                raise KapitanError("parameters.kapitan.secrets not defined in {}".format(args.target_name))

            recipients = kap_inv_params['secrets']['gpg']['recipients']
        if not recipients:
            raise KapitanError("No GPG recipients specified. Use --recipients or specify them in " +
                               "parameters.kapitan.secrets.gpg.recipients and use --target")
        type_name, token_path = token_name.split(":")
        tag = '?{{gpg:{}}}'.format(token_path)
        secret_obj = ref_controller[tag]
        secret_obj.update_recipients(recipients)
        ref_controller[tag] = secret_obj

    elif token_name.startswith("gkms:"):
        key = args.key
        if args.target_name:
            inv = inventory_reclass(args.inventory_path)
            kap_inv_params = inv['nodes'][args.target_name]['parameters']['kapitan']
            if 'secrets' not in kap_inv_params:
                raise KapitanError("parameters.kapitan.secrets not defined in {}".format(args.target_name))

            key = kap_inv_params['secrets']['gkms']['key']
        if not key:
            raise KapitanError("No KMS key specified. Use --key or specify it in parameters.kapitan.secrets.gkms.key and use --target")
        type_name, token_path = token_name.split(":")
        tag = '?{{gkms:{}}}'.format(token_path)
        secret_obj = ref_controller[tag]
        secret_obj.update_key(key)
        ref_controller[tag] = secret_obj

    elif token_name.startswith("awskms:"):
        key = args.key
        if args.target_name:
            inv = inventory_reclass(args.inventory_path)
            kap_inv_params = inv['nodes'][args.target_name]['parameters']['kapitan']
            if 'secrets' not in kap_inv_params:
                raise KapitanError("parameters.kapitan.secrets not defined in {}".format(args.target_name))

            key = kap_inv_params['secrets']['awskms']['key']
        if not key:
            raise KapitanError("No KMS key specified. Use --key or specify it in parameters.kapitan.secrets.awskms.key and use --target")
        type_name, token_path = token_name.split(":")
        tag = '?{{awskms:{}}}'.format(token_path)
        secret_obj = ref_controller[tag]
        secret_obj.update_key(key)
        ref_controller[tag] = secret_obj

    else:
        fatal_error("Invalid token: {name}. Try using gpg/gkms/awskms:{name}".format(name=token_name))


def ref_reveal(args, ref_controller):
    "Reveal secrets in file_name"
    revealer = Revealer(ref_controller)
    file_name = args.file
    if file_name is None:
        fatal_error('--file is required with --reveal')
    try:
        if file_name == '-':
            out = revealer.reveal_raw_file(None)
            sys.stdout.write(out)
        elif file_name:
            for rev_obj in revealer.reveal_path(file_name):
                sys.stdout.write(rev_obj.content)
    except (RefHashMismatchError, KeyError):
        raise KapitanError("Reveal failed for file {name}".format(name=file_name))


def secret_update_validate(args, ref_controller):
    "Validate and/or update target secrets"
    # update gpg recipients/gkms/awskms key for all secrets in secrets_path
    # use --refs-path to set scanning path
    inv = inventory_reclass(args.inventory_path)
    targets = set(inv['nodes'].keys())
    secrets_path = os.path.abspath(args.refs_path)
    target_token_paths = search_target_token_paths(secrets_path, targets)
    ret_code = 0

    for target_name, token_paths in target_token_paths.items():
        kap_inv_params = inv['nodes'][target_name]['parameters']['kapitan']
        if 'secrets' not in kap_inv_params:
            raise KapitanError("parameters.kapitan.secrets not defined in {}".format(target_name))

        try:
            recipients = kap_inv_params['secrets']['gpg']['recipients']
        except KeyError:
            recipients = None
        try:
            gkey = kap_inv_params['secrets']['gkms']['key']
        except KeyError:
            gkey = None
        try:
            awskey = kap_inv_params['secrets']['awskms']['key']
        except KeyError:
            awskey = None
        try:
            vaultkv = kap_inv_params['secrets']['vaultkv']['auth']
        except KeyError:
            vaultkv = None

        for token_path in token_paths:
            if token_path.startswith("?{gpg:"):
                if not recipients:
                    logger.debug("secret_update_validate: target: %s has no inventory gpg recipients, skipping %s", target_name, token_path)
                    continue
                secret_obj = ref_controller[token_path]
                target_fingerprints = set(lookup_fingerprints(recipients))
                secret_fingerprints = set(lookup_fingerprints(secret_obj.recipients))
                if target_fingerprints != secret_fingerprints:
                    if args.validate_targets:
                        logger.info("%s recipient mismatch", token_path)
                        to_remove = secret_fingerprints.difference(target_fingerprints)
                        to_add = target_fingerprints.difference(secret_fingerprints)
                        if to_remove:
                            logger.info("%s needs removal", to_remove)
                        if to_add:
                            logger.info("%s needs addition", to_add)
                        ret_code = 1
                    else:
                        new_recipients = [dict([("fingerprint", f), ]) for f in target_fingerprints]
                        secret_obj.update_recipients(new_recipients)
                        ref_controller[token_path] = secret_obj

            elif token_path.startswith("?{gkms:"):
                if not gkey:
                    logger.debug("secret_update_validate: target: %s has no inventory gkms key, skipping %s", target_name, token_path)
                    continue
                secret_obj = ref_controller[token_path]
                if gkey != secret_obj.key:
                    if args.validate_targets:
                        logger.info("%s key mismatch", token_path)
                        ret_code = 1
                    else:
                        secret_obj.update_key(gkey)
                        ref_controller[token_path] = secret_obj

            elif token_path.startswith("?{awskms:"):
                if not awskey:
                    logger.debug("secret_update_validate: target: %s has no inventory awskms key, skipping %s", target_name, token_path)
                    continue
                secret_obj = ref_controller[token_path]
                if awskey != secret_obj.key:
                    if args.validate_targets:
                        logger.info("%s key mismatch", token_path)
                        ret_code = 1
                    else:
                        secret_obj.update_key(awskey)
                        ref_controller[token_path] = secret_obj

            elif token_path.startswith("?{vaultkv:"):
                if not vaultkv:
                    logger.debug("secret_update_validate: target: %s has no inventory vaultkv parameter 'auth', skipping %s", target_name, token_path)
                    continue
                secret_obj = ref_controller[token_path]

            else:
                logger.info("Invalid secret %s, could not get type, skipping", token_path)
                ret_code = 1

    sys.exit(ret_code)



# Copyright 2019 The Kapitan Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"hashicorp vault kv secrets module"

import base64
import logging
import os
from binascii import Error as b_error
from sys import exit

from kapitan import cached
from kapitan.errors import KapitanError
from kapitan.refs.base import RefError
from kapitan.refs.base64 import Base64Ref, Base64RefBackend

import hvac
from hvac.exceptions import Forbidden, InvalidPath

logger = logging.getLogger(__name__)


class VaultError(KapitanError):
    """Generic vault errors"""
    pass


def get_env(parameter):
    """
    The following variables need to be exported to the environment or defined in inventory.
        * VAULT_ADDR: url for vault
        * VAULT_SKIP_VERIFY=true: if set, do not verify presented TLS certificate before communicating with Vault server.
        * VAULT_CLIENT_KEY: path to an unencrypted PEM-encoded private key matching the client certificate
        * VAULT_CLIENT_CERT: path to a PEM-encoded client certificate for TLS authentication to the Vault server
        * VAULT_CACERT: path to a PEM-encoded CA cert file to use to verify the Vault server TLS certificate
        * VAULT_CAPATH: path to a directory of PEM-encoded CA cert files to verify the Vault server TLS certificate
        * VAULT_NAMESPACE: specify the Vault Namespace, if you have one

    Following keys are used to creates a new hvac client instance.
        :param url: Base URL for the Vault instance being addressed.
        :type url: str
        :param cert: Certificates for use in requests sent to the Vault instance. This should be a tuple with the
            certificate and then key.
        :type cert: tuple
        :param verify: Either a boolean to indicate whether TLS verification should be performed when sending requests to Vault,
            or a string pointing at the CA bundle to use for verification.
            See http://docs.python-requests.org/en/master/user/advanced/#ssl-cert-verification.
        :type verify: Union[bool,str]
        :param namespace: Optional Vault Namespace.
        :type namespace: str
    """
    client_parameters = {}
    client_parameters['url'] = parameter.get('VAULT_ADDR',
                                             os.getenv('VAULT_ADDR', default=''))
    client_parameters['namespace'] = parameter.get('VAULT_NAMESPACE',
                                             os.getenv('VAULT_NAMESPACE', default=''))
    # VERIFY VAULT SERVER TLS CERTIFICATE
    skip_verify = parameter.get('VAULT_SKIP_VERIFY', os.getenv('VAULT_SKIP_VERIFY', default=''))

    if skip_verify.lower() == 'false':
        cert = parameter.get('VAULT_CACERT', os.getenv('VAULT_CACERT', default=''))
        if not cert:
            cert_path = parameter.get('VAULT_CAPATH', os.getenv('VAULT_CAPATH', default=''))
            if not cert_path:
                raise Exception('Neither VAULT_CACERT nor VAULT_CAPATH specified')
            client_parameters['verify'] = cert_path
        else:
            client_parameters['verify'] = cert
    else:
        client_parameters['verify'] = False

    # CLIENT CERTIFICATE FOR TLS AUTHENTICATION
    client_key = parameter.get('VAULT_CLIENT_KEY', os.getenv('VAULT_CLIENT_KEY', default=''))
    client_cert = parameter.get('VAULT_CLIENT_CERT', os.getenv('VAULT_CLIENT_CERT', default=''))
    if client_key != '' and client_cert != '':
        client_parameters['cert'] = (client_cert, client_key)
    return client_parameters

def vault_obj(vault_parameters):
    """
    vault_parameters: necessary parameters to authenticate & get value from vault, provided by inventory
    e.g.:
        auth: userpass
        VAULT_ADDR: http://127.0.0.1:8200
        VAULT_SKIP_VERIFY: false
    Authenticate client to server and return client object
    """
    env = get_env(vault_parameters)

    client = hvac.Client(
        **{k: v for k, v in env.items() if k is not 'auth'}
    )

    auth_type = vault_parameters['auth']
    # GET TOKEN EITHER FROM ENVIRONMENT OF FILE
    if auth_type in ['token', 'github']:
        env['token'] = os.getenv('VAULT_TOKEN')
        if not env['token']:
            try:
                token_file = os.path.join(os.path.expanduser('~'), '.vault-token')
                with open(token_file, 'r') as f:
                    env['token'] = f.read()
                if env['token'] == '':
                    raise VaultError('{file} is empty'.format(file=token_file))
            except IOError:
                VaultError("Cannot read file {file}".format(file=token_file))
    # DIFFERENT LOGIN METHOD BASED ON AUTHENTICATION TYPE
    if auth_type == 'token':
        client.token = env['token']
    elif auth_type == 'ldap':
        client.auth.ldap.login(
            username=os.getenv('VAULT_USERNAME'),
            password=os.getenv('VAULT_PASSWORD')
        )
    elif auth_type == 'userpass':
        client.auth_userpass(
            username=os.getenv('VAULT_USERNAME'),
            password=os.getenv('VAULT_PASSWORD')
        )
    elif auth_type == 'approle':
        client.auth_approle(
            os.getenv('VAULT_ROLE_ID'),
            secret_id=os.getenv('VAULT_SECRET_ID')
        )
    elif auth_type == 'github':
        client.auth.github.login(token=env['token'])
    else:
        raise "Authentication type '{auth}' not supported".format(auth=auth_type)

    if client.is_authenticated():
        return client
    else:
        VaultError("Vault Authentication Error, Environment Variables defined?")


class VaultSecret(Base64Ref):
    """
    Hashicorp Vault support for KV Secret Engine
    """
    def __init__(self, data, vault_params, **kwargs):
        """
        Set vault parameter and encoding of data
        """
        self.data = data
        self.vault_params = vault_params
        super().__init__(self.data, **kwargs)
        self.type_name = 'vaultkv'

    @classmethod
    def from_params(cls, data, ref_params):
        """
        Return new VaultSecret from data and ref_params: target_name
        parameters will be grabbed from the inventory via target_name
        """
        try:
            target_name = ref_params.kwargs['target_name']
            if target_name is None:
                raise ValueError('target_name not set')

            target_inv = cached.inv['nodes'].get(target_name, None)
            if target_inv is None:
                raise ValueError('target_inv not set')

            ref_params.kwargs['vault_params'] = target_inv['parameters']['kapitan']['secrets']['vaultkv']
            return cls(data, **ref_params.kwargs)
        except KeyError:
            raise RefError("Could not create VaultSecret: vaultkv parameters missing")

    @classmethod
    def from_path(cls, ref_full_path, **kwargs):
        return super().from_path(ref_full_path, encrypt=False)

    def reveal(self):
        """
        Returns decrypted data
        """
        # can't use super().reveal() as we want bytes
        try:
            self.data = base64.b64decode(self.data, validate=True)
        except b_error:
            exit("non-alphabet characters in the data")

        return self._decrypt()

    def _decrypt(self):
        """
        Authenticate with Vault server & returns value of the key from secret

        :returns: secret in plain text
        """
        try:
            client = vault_obj(self.vault_params)
            # token will comprise of two parts path_in_vault:key
            data = self.data.decode('utf-8').rstrip().split(':')
            return_data = ''
            if self.vault_params.get('engine') == 'kv':
                response = client.secrets.kv.v1.read_secret(path=data[0],
                                                            mount_point=self.vault_params.get('mount', 'secret'))
                return_data = response['data'][data[1]]
            else:
                response = client.secrets.kv.v2.read_secret_version(path=data[0],
                                                                    mount_point=self.vault_params.get('mount', 'secret'))
                return_data = response['data']['data'][data[1]]
            client.adapter.close()
        except Forbidden:
            VaultError(
                'Permission Denied. '+
                'make sure the token is authorised to access {path} on Vault'.format( path=data[0])
            )
        except InvalidPath:
            VaultError(
                '{path} does not exist on Vault secret'.format(path=data[0])
            )

        if return_data is '':
            VaultError(
                "'{key}' doesn't exist on '{path}'".format(key=data[1], path=data[0])
            )
        return return_data

    def dump(self):
        """
        Returns dict with keys/values to be serialised.
        """
        return {"data": self.data, "encoding": self.encoding,
                "type": self.type_name, "vault_params": self.vault_params}

class VaultBackend(Base64RefBackend):
    def __init__(self, path, ref_type=VaultSecret):
        "init VaultBackend ref backend type"
        super().__init__(path, ref_type)
        self.type_name = 'vaultkv'



#!/usr/bin/env python3
#
# Copyright 2019 The Kapitan Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"vault secrets tests"

import os
import shutil
import tempfile
import unittest
from time import sleep

import docker
from kapitan.refs.base import RefController, Revealer
from kapitan.refs.secrets.vaultkv import VaultSecret, vault_obj

import hvac

# Create temporary folder
REFS_HOME = tempfile.mkdtemp()
REF_CONTROLLER = RefController(REFS_HOME)
REVEALER = Revealer(REF_CONTROLLER)

# Create Vault docker container
client = docker.from_env()
env = {'VAULT_LOCAL_CONFIG': '{"backend": {"file": {"path": "/vault/file"}}, "listener":{"tcp":{"address":"0.0.0.0:8200","tls_disable":"true"}}}'}
vault_container = client.containers.run(image='vault', cap_add=['IPC_LOCK'], ports={'8200': '8200'},
                                       environment=env, detach=True, remove=True, command='server')

# make sure the container is up & running before testing
while vault_container.status != "running":
    sleep(2)
    vault_container.reload()

class VaultSecretTest(unittest.TestCase):
    "Test Vault Secret"

    @classmethod
    def setUpClass(cls):
        # Initialize vault, unseal, mount secret engine & add auth
        cls.client = hvac.Client()
        init = cls.client.sys.initialize()
        cls.client.sys.submit_unseal_keys(init['keys'])
        os.environ['VAULT_ROOT_TOKEN'] = init['root_token']
        cls.client.adapter.close()
        cls.client = hvac.Client(token=init['root_token'])
        cls.client.sys.enable_secrets_engine(backend_type='kv-v2', path='secret')
        test_policy = '''
        path "secret/*" {
          capabilities = ["read", "list"]
        }
        '''
        policy = 'test_policy'
        cls.client.sys.create_or_update_policy(name=policy, policy=test_policy)
        os.environ['VAULT_USERNAME'] = 'test_user'
        os.environ['VAULT_PASSWORD'] = 'test_password'
        cls.client.sys.enable_auth_method('userpass')
        cls.client.create_userpass(username='test_user', password='test_password', policies=[policy])
        cls.client.sys.enable_auth_method('approle')
        cls.client.create_role('test_role')
        os.environ['VAULT_ROLE_ID'] = cls.client.get_role_id('test_role')
        os.environ['VAULT_SECRET_ID'] = cls.client.create_role_secret_id('test_role')['data']['secret_id']
        os.environ['VAULT_TOKEN'] = cls.client.create_token(policies=[policy], lease='1h')['auth']['client_token']

    @classmethod
    def tearDownClass(cls):
        cls.client.adapter.close()
        vault_container.stop()
        shutil.rmtree(REFS_HOME, ignore_errors=True)
        for i in ['ROOT_TOKEN', 'TOKEN', 'USERNAME', 'PASSWORD', 'ROLE_ID', 'SECRET_ID']:
            del os.environ['VAULT_' + i]

    def test_token_authentication(self):
        '''
        Authenticate using token
        '''
        parameters = {'auth': 'token'}
        test_client = vault_obj(parameters)
        self.assertTrue(test_client.is_authenticated(),
                        msg='Authentication with token failed')
        test_client.adapter.close()

    def test_userpss_authentication(self):
        '''
        Authenticate using userpass
        '''
        parameters = {'auth': 'userpass'}
        test_client = vault_obj(parameters)
        self.assertTrue(test_client.is_authenticated(),
                        msg='Authentication with userpass failed')
        test_client.adapter.close()

    def test_approle_authentication(self):
        '''
        Authenticate using approle
        '''
        parameters = {'auth': 'approle'}
        test_client = vault_obj(parameters)
        self.assertTrue(test_client.is_authenticated(),
                        msg='Authentication with approle failed')
        test_client.adapter.close()

    def test_vault_write_reveal(self):
        '''
        Write secret, confirm secret file exists, reveal and compare content
        '''
        tag = '?{vaultkv:secret/batman}'
        secret = {
            'some_random_value': 'something_secret'
        }
        self.client.secrets.kv.v2.create_or_update_secret(
            path='foo',
            secret=secret,
        )
        env = {'auth': 'token'}
        file_data = "foo:some_random_value".encode()
        REF_CONTROLLER[tag] = VaultSecret(file_data, env)

        # confirming secret file exists
        self.assertTrue(os.path.isfile(os.path.join(REFS_HOME, 'secret/batman')),
                        msg="Secret file doesn't exist")
        file_with_secret_tags = tempfile.mktemp()
        with open(file_with_secret_tags, 'w') as fp:
            fp.write('File contents revealed: {}'.format(tag))
        revealed = REVEALER.reveal_raw_file(file_with_secret_tags)

        # confirming secerts are correctly revealed
        self.assertEqual(
            "File contents revealed: {}".format(secret['some_random_value']), revealed
        )



#!/usr/bin/env python3
#
# Copyright 2019 The Kapitan Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging
import os

from kapitan.inputs.base import InputType, CompiledFile
from kapitan.resources import inventory
from kapitan.utils import render_jinja2

logger = logging.getLogger(__name__)


class Jinja2(InputType):
    def __init__(self, compile_path, search_paths, ref_controller):
        super().__init__("jinja2", compile_path, search_paths, ref_controller)

    def compile_file(self, file_path, compile_path, ext_vars, **kwargs):
        """
        Write items in path as jinja2 rendered files to compile_path.
        path can be either a file or directory.
        kwargs:
            reveal: default False, set to reveal refs on compile
            target_name: default None, set to current target being compiled
        """
        reveal = kwargs.get('reveal', False)
        target_name = kwargs.get('target_name', None)

        # set ext_vars and inventory for jinja2 context
        context = ext_vars.copy()
        context["inventory"] = inventory(self.search_paths, target_name)
        context["inventory_global"] = inventory(self.search_paths, None)
        jinja2_filters = kwargs.get('jinja2_filters')

        for item_key, item_value in render_jinja2(file_path, context,
                                                  jinja2_filters=jinja2_filters).items():
            full_item_path = os.path.join(compile_path, item_key)
            os.makedirs(os.path.dirname(full_item_path), exist_ok=True)

            with CompiledFile(full_item_path, self.ref_controller, mode="w", reveal=reveal,
                              target_name=target_name) as fp:
                fp.write(item_value["content"])
                mode = item_value["mode"]
                os.chmod(full_item_path, mode)
                logger.debug("Wrote %s with mode %.4o", full_item_path, mode)

    def default_output_type(self):
        # no output_type options for jinja2
        return None



import hashlib
import logging
import os
import re
import tarfile
import tempfile
from collections import defaultdict
from distutils.dir_util import copy_tree
from functools import partial
from shutil import copyfile
from zipfile import ZipFile

from kapitan.errors import GitSubdirNotFoundError
from kapitan.utils import make_request

from git import Repo

logger = logging.getLogger(__name__)

# this is used just to change the root directory of save_dir in testing
DEPENDENCY_OUTPUT_CONFIG = {}


def fetch_dependencies(target_objs, pool):
    """
    parses through the dependencies parameters in target_objs and fetches them
    all dependencies are first fetched into tmp dir, after which they are copied to their respective output_path.
    dependencies whose output_path already exists are skipped
    """
    temp_dir = tempfile.mkdtemp()
    # there could be multiple dependency items per source_uri due to reclass inheritance or
    # other user requirements. So create a mapping from source_uri to a set of dependencies with
    # that source_uri
    git_deps = defaultdict(list)
    http_deps = defaultdict(list)

    # this dict is to make sure no duplicated output_paths exist per source
    deps_output_paths = defaultdict(set)
    for target_obj in target_objs:
        try:
            dependencies = target_obj["dependencies"]
            for item in dependencies:
                dependency_type = item["type"]
                source_uri = item["source"]
                if "root_dir" in DEPENDENCY_OUTPUT_CONFIG:
                    # should only be used in test environment
                    item["output_path"] = os.path.join(DEPENDENCY_OUTPUT_CONFIG["root_dir"], item["output_path"])
                output_path = item["output_path"]
                if os.path.exists(os.path.abspath(output_path)) and not item.get('unpack', False):
                    # if unpack: True, then allow the user to specify the parent directory as output_path where the
                    # files will be extracted to
                    logger.info("Dependency {} : already exists. Ignoring".format(output_path))
                    continue

                if output_path in deps_output_paths[source_uri]:
                    # if the output_path is duplicated for the same source_uri
                    continue
                else:
                    deps_output_paths[source_uri].add(output_path)

                if dependency_type == "git":
                    git_deps[source_uri].append(item)
                elif dependency_type in ("http", "https"):
                    http_deps[source_uri].append(item)

        except KeyError:
            continue

    git_worker = partial(fetch_git_dependency, save_dir=temp_dir)
    http_worker = partial(fetch_http_dependency, save_dir=temp_dir)
    [p.get() for p in pool.imap_unordered(http_worker, http_deps.items()) if p]
    [p.get() for p in pool.imap_unordered(git_worker, git_deps.items()) if p]


def fetch_git_dependency(dep_mapping, save_dir):
    """
    fetches a git repository at source into save_dir, and copy the repository into output_path
    ref is used to checkout if exists
    only subdir is copied into output_path if specified
    """
    source, deps = dep_mapping
    fetch_git_source(source, save_dir)
    for dep in deps:
        repo_path = os.path.join(save_dir, os.path.basename(source))
        repo = Repo(repo_path)
        output_path = dep["output_path"]
        copy_src_path = repo_path
        if 'ref' in dep:
            ref = dep['ref']
            repo.git.checkout(ref)
        else:
            repo = Repo(repo_path)
            repo.git.checkout("master")  # default ref

        if 'subdir' in dep:
            sub_dir = dep['subdir']
            full_subdir = os.path.join(repo_path, sub_dir)
            if os.path.isdir(full_subdir):
                copy_src_path = full_subdir
            else:
                raise GitSubdirNotFoundError("Dependency {} : subdir {} not found in repo".format(source, sub_dir))

        if not os.path.exists(os.path.abspath(output_path)):
            copy_tree(copy_src_path, output_path)
            logger.info("Dependency {} : saved to {}".format(source, output_path))


def fetch_git_source(source, save_dir):
    """clones a git repository at source and saves into save_dir"""
    logger.info("Dependency {} : fetching now".format(source))
    Repo.clone_from(source, os.path.join(save_dir, os.path.basename(source)))
    logger.info('Dependency {} : successfully fetched'.format(source))


def fetch_http_dependency(dep_mapping, save_dir):
    """
    fetches a http[s] file at source and saves into save_dir, after which it is copied into
    the output_path stored in dep_mapping
    """
    source, deps = dep_mapping
    content_type = fetch_http_source(source, save_dir)
    path_hash = hashlib.sha256(os.path.dirname(source).encode()).hexdigest()[:8]
    copy_src_path = os.path.join(save_dir, path_hash + os.path.basename(source))
    for dep in deps:
        output_path = dep["output_path"]
        if dep.get('unpack', False):
            # ensure that the directory we are extracting to exists
            os.makedirs(output_path, exist_ok=True)
            is_unpacked = False
            if content_type == 'application/x-tar':
                tar = tarfile.open(copy_src_path)
                tar.extractall(path=output_path)
                tar.close()
                is_unpacked = True
            elif content_type == 'application/zip':
                zfile = ZipFile(copy_src_path)
                zfile.extractall(output_path)
                zfile.close()
                is_unpacked = True
            elif content_type in ['application/octet-stream', 'application/x-gzip']:
                if re.search(r'(\.tar\.gz|\.tgz)$', source):
                    tar = tarfile.open(copy_src_path)
                    tar.extractall(path=output_path)
                    tar.close()
                    is_unpacked = True
            if is_unpacked:
                logger.info("Dependency {} : extracted to {}".format(source, output_path))
            else:
                logger.info("Dependency {} : Content-Type {} is not supported for unpack. Ignoring save".
                            format(source, content_type))
        else:
            # we are downloading a single file
            parent_dir = os.path.dirname(output_path)
            if parent_dir != '':
                os.makedirs(parent_dir, exist_ok=True)
            copyfile(copy_src_path, output_path)
            logger.info("Dependency {} : saved to {}".format(source, output_path))


def fetch_http_source(source, save_dir):
    """downloads a http[s] file from source and saves into save_dir"""
    logger.info("Dependency {} : fetching now".format(source))
    content, content_type = make_request(source)
    logger.info("Dependency {} : successfully fetched".format(source))
    if content is not None:
        basename = os.path.basename(source)
        # to avoid collisions between basename(source)
        path_hash = hashlib.sha256(os.path.dirname(source).encode()).hexdigest()[:8]
        full_save_path = os.path.join(save_dir, path_hash + basename)
        with open(full_save_path, 'wb') as f:
            f.write(content)
        return content_type
    else:
        logger.warning("Dependency {} : failed to fetch".format(source))
        return None



# Copyright 2019 The Kapitan Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"binary tests"

import base64
import contextlib
from distutils import dir_util
import io
import os
import tempfile
import shutil
import subprocess
import sys
import unittest

from kapitan.cached import reset_cache
from kapitan.cli import main
from kapitan.utils import directory_hash

SECRETS_PATH = tempfile.mkdtemp()
BINARY_PATH = 'dist/kapitan'


@unittest.skipIf(not os.path.exists(BINARY_PATH), 'kapitan binary not found')
class BinaryTest(unittest.TestCase):
    def test_cli_secret_validate_targets(self):
        """
        run $ kapitan secrets --validate-targets
        expect 0 (success) exit status code
        """
        argv = ["kapitan", "secrets", "--validate-targets",
                "--secrets-path", "examples/kubernetes/secrets/targets/",
                "--inventory-path", "examples/kubernetes/inventory/"]
        with self.assertRaises(SystemExit) as cm:
            sys.argv = argv
            main()
        argv[0] = BINARY_PATH
        result = subprocess.run(argv, stdout=subprocess.PIPE)
        self.assertEqual(cm.exception.code, result.returncode)

    def test_cli_secret_write_reveal_gkms(self):
        """
        run $ kapitan secrets --write gkms:test_secret
        and $ kapitan secrets --reveal
        using mock KMS key
        """
        test_secret_content = "mock"
        test_secret_file = tempfile.mktemp()
        with open(test_secret_file, "w") as fp:
            fp.write(test_secret_content)

        argv = [BINARY_PATH, "secrets", "--write", "gkms:test_secret",
                "-f", test_secret_file,
                "--secrets-path", SECRETS_PATH,
                "--key", "mock"]
        subprocess.run(argv)

        test_tag_content = "revealing: ?{gkms:test_secret}"
        test_tag_file = tempfile.mktemp()
        with open(test_tag_file, "w") as fp:
            fp.write(test_tag_content)
        argv = [BINARY_PATH, "secrets", "--reveal",
                "-f", test_tag_file,
                "--secrets-path", SECRETS_PATH]

        result = subprocess.run(argv, stdout=subprocess.PIPE)
        self.assertEqual("revealing: {}".format(test_secret_content),
                         result.stdout.decode('utf-8'))

        os.remove(test_tag_file)

    def test_cli_secret_write_reveal_awskms(self):
        """
        run $ kapitan secrets --write awskms:test_secret
        and $ kapitan secrets --reveal
        using mock KMS key
        """
        test_secret_content = "mock"
        test_secret_file = tempfile.mktemp()
        with open(test_secret_file, "w") as fp:
            fp.write(test_secret_content)

        argv = [BINARY_PATH, "secrets", "--write", "awskms:test_secret",
                "-f", test_secret_file,
                "--secrets-path", SECRETS_PATH,
                "--key", "mock"]
        subprocess.run(argv)

        test_tag_content = "revealing: ?{awskms:test_secret}"
        test_tag_file = tempfile.mktemp()
        with open(test_tag_file, "w") as fp:
            fp.write(test_tag_content)
        argv = [BINARY_PATH, "secrets", "--reveal",
                "-f", test_tag_file,
                "--secrets-path", SECRETS_PATH]
        result = subprocess.run(argv, stdout=subprocess.PIPE)
        self.assertEqual("revealing: {}".format(test_secret_content),
                         result.stdout.decode('utf-8'))

        os.remove(test_tag_file)

    def test_cli_secret_write_ref(self):
        """
        run $ kapitan secrets --write ref:test_secret
        and $ kapitan secrets --reveal -f sometest_file
        """
        test_secret_content = "secret_value!"
        test_secret_file = tempfile.mktemp()
        with open(test_secret_file, "w") as fp:
            fp.write(test_secret_content)

        argv = [BINARY_PATH, "secrets", "--write", "ref:test_secret",
                "-f", test_secret_file,
                "--secrets-path", SECRETS_PATH]
        subprocess.run(argv)
        test_tag_content = "revealing: ?{ref:test_secret}"
        test_tag_file = tempfile.mktemp()
        with open(test_tag_file, "w") as fp:
            fp.write(test_tag_content)
        argv = [BINARY_PATH, "secrets", "--reveal",
                "-f", test_tag_file,
                "--secrets-path", SECRETS_PATH]
        result = subprocess.run(argv, stdout=subprocess.PIPE)
        self.assertEqual("revealing: {}".format(test_secret_content),
                         result.stdout.decode('utf-8'))

        os.remove(test_tag_file)

    def test_cli_secret_write_base64_ref(self):
        """
        run $ kapitan secrets --write ref:test_secret --base64
        and $ kapitan secrets --reveal -f sometest_file
        """
        test_secret_content = "secret_value!"
        test_secret_content_b64 = base64.b64encode(test_secret_content.encode())
        test_secret_file = tempfile.mktemp()
        with open(test_secret_file, "w") as fp:
            fp.write(test_secret_content)

        argv = [BINARY_PATH, "secrets", "--write", "ref:test_secret",
                "--base64", "-f", test_secret_file,
                "--secrets-path", SECRETS_PATH]
        subprocess.run(argv)

        test_tag_content = "revealing: ?{ref:test_secret}"
        test_tag_file = tempfile.mktemp()
        with open(test_tag_file, "w") as fp:
            fp.write(test_tag_content)
        argv = [BINARY_PATH, "secrets", "--reveal",
                "-f", test_tag_file,
                "--secrets-path", SECRETS_PATH]

        result = subprocess.run(argv, stdout=subprocess.PIPE)
        self.assertEqual("revealing: {}".format(test_secret_content_b64.decode()),
                         result.stdout.decode('utf-8'))

        os.remove(test_tag_file)

    def test_cli_secret_subvar_ref(self):
        """
        run $ kapitan secrets --write ref:test_secret
        and $ kapitan secrets --reveal -f sometest_file
        """
        test_secret_content = """
        var1:
          var2: hello
        var3:
          var4: world
        """
        test_secret_file = tempfile.mktemp()
        with open(test_secret_file, "w") as fp:
            fp.write(test_secret_content)

        argv = [BINARY_PATH, "secrets", "--write", "ref:test_secret_subvar",
                "-f", test_secret_file,
                "--secrets-path", SECRETS_PATH]
        subprocess.run(argv)
        test_tag_content = """
        revealing1: ?{ref:test_secret_subvar@var1.var2}
        revealing2: ?{ref:test_secret_subvar@var3.var4}
        """
        test_tag_file = tempfile.mktemp()
        with open(test_tag_file, "w") as fp:
            fp.write(test_tag_content)
        argv = [BINARY_PATH, "secrets", "--reveal",
                "-f", test_tag_file,
                "--secrets-path", SECRETS_PATH]

        result = subprocess.run(argv, stdout=subprocess.PIPE)
        expected = """
        revealing1: {}
        revealing2: {}
        """
        self.assertEqual(expected.format("hello", "world"), result.stdout.decode('utf-8'))
        os.remove(test_tag_file)

    def test_cli_searchvar(self):
        """
        run $ kapitan searchvar mysql.replicas
        """
        argv = ["kapitan", "searchvar", "mysql.replicas",
                "--inventory-path", "examples/kubernetes/inventory/"]

        # set stdout as string
        stdout = io.StringIO()
        with contextlib.redirect_stdout(stdout):
            sys.argv = argv
            main()

        argv[0] = BINARY_PATH
        result = subprocess.run(argv, stdout=subprocess.PIPE)

        self.assertEqual(result.stdout.decode('utf-8'), stdout.getvalue())

    def test_cli_inventory(self):
        """
        run $ kapitan inventory -t minikube-es -F -p cluster
        """
        argv = ["kapitan", "inventory", "-t", "minikube-es", "-F", "-p", "cluster",
                "--inventory-path", "examples/kubernetes/inventory/"]

        # set stdout as string
        stdout = io.StringIO()
        with contextlib.redirect_stdout(stdout):
            sys.argv = argv
            main()
        argv[0] = BINARY_PATH
        result = subprocess.run(argv, stdout=subprocess.PIPE)
        self.assertEqual(result.stdout.decode('utf-8'),
                         stdout.getvalue())

    def tearDown(self):
        shutil.rmtree(SECRETS_PATH, ignore_errors=True)


@unittest.skipIf(not os.path.exists(BINARY_PATH), 'kapitan binary not found')
class TestTerraformCompile(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        os.chdir(os.getcwd() + '/examples/terraform/')

    def test_cli_terraform_compile(self):
        """
        run $kapitan compile
        """
        sys.argv = ["kapitan", "compile"]
        main()
        compiled_dir_hash = directory_hash(os.getcwd() + '/compiled')
        test_compiled_dir_hash = directory_hash(
            os.getcwd() + '/../../tests/test_terraform_compiled')
        self.assertEqual(compiled_dir_hash, test_compiled_dir_hash)
        compile_path = tempfile.mkdtemp()
        argv = ["kapitan", "compile"]
        sys.argv = argv
        with contextlib.redirect_stdout(io.StringIO()):
            main()
        argv[0] = '../../' + BINARY_PATH
        argv.extend(["--output-path", compile_path])
        subprocess.run(argv, stdout=subprocess.DEVNULL)
        main_compiled_dir_hash = directory_hash(os.getcwd() + '/compiled')
        binary_compiled_dir_hash = directory_hash(compile_path + '/compiled')
        self.assertEqual(main_compiled_dir_hash, binary_compiled_dir_hash)

    @classmethod
    def tearDownClass(cls):
        os.chdir(os.getcwd() + '/../../')
        reset_cache()


@unittest.skipIf(not os.path.exists(BINARY_PATH), 'kapitan binary not found')
class TestKubernetesCompile(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        os.chdir(os.getcwd() + '/examples/kubernetes/')

    def test_cli_kubenetes_compile(self):
        """
        run $kapitan compile
        """
        compile_path = tempfile.mkdtemp()
        argv = ["kapitan", "compile"]
        sys.argv = argv
        with contextlib.redirect_stdout(io.StringIO()):
            main()
        argv[0] = '../../' + BINARY_PATH
        argv.extend(["--output-path", compile_path])
        subprocess.run(argv, stdout=subprocess.DEVNULL)
        main_compiled_dir_hash = directory_hash(os.getcwd() + '/compiled')
        binary_compiled_dir_hash = directory_hash(compile_path + '/compiled')
        self.assertEqual(main_compiled_dir_hash, binary_compiled_dir_hash)

    @classmethod
    def tearDownClass(cls):
        os.chdir(os.getcwd() + '/../../')
        reset_cache()


@unittest.skipIf(not os.path.exists(BINARY_PATH), 'kapitan binary not found')
class TestDependencyManager(unittest.TestCase):
    def test_compile_fetch(self):
        temp = tempfile.mkdtemp()
        dir_util.copy_tree('tests/test_resources', temp)
        argv = [os.path.join(os.getcwd(), BINARY_PATH), "compile", "--fetch",
                "--output-path", temp, "-t", "nginx", "nginx-dev", "-p", "4"]
        subprocess.run(argv, cwd=temp, stdout=subprocess.DEVNULL)
        self.assertTrue(os.path.isdir(os.path.join(temp, "components", "tests")))
        self.assertTrue(os.path.isdir(os.path.join(temp, "components", "acs-engine-autoscaler")))
        self.assertTrue(os.path.isdir(os.path.join(temp, "components", "kapitan-repository")))
        self.assertTrue(os.path.isdir(os.path.join(temp, "components", "source")))

    def tearDown(self):
        reset_cache()


@unittest.skipIf(not os.path.exists(BINARY_PATH), 'kapitan binary not found')
class TestHelmInputBinary(unittest.TestCase):
    def setUp(self):
        os.chdir(os.path.join("tests", "test_resources"))

    def test_compile_helm_input(self):
        """compile targets with helm inputs in parallel"""
        temp_main = tempfile.mkdtemp()
        sys.argv = ["kapitan", "compile", "--output-path", temp_main, '-t', 'nginx-ingress', 'nginx-ingress-helm-params', 'acs-engine-autoscaler']
        with contextlib.redirect_stdout(io.StringIO()):
            main()
        main_hash = directory_hash(temp_main + '/compiled')

        temp_bin = tempfile.mkdtemp()
        subprocess.run(['../../' + BINARY_PATH, "compile", "--output-path", temp_bin, '-t', 'nginx-ingress', 'nginx-ingress-helm-params', 'acs-engine-autoscaler'], stdout=subprocess.DEVNULL)
        bin_hash = directory_hash(temp_bin + '/compiled')
        self.assertEqual(bin_hash, main_hash)

    @classmethod
    def tearDownClass(cls):
        reset_cache()
        os.chdir('../../')



    #!/usr/bin/env python3
#
# Copyright 2019 The Kapitan Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"helm input tests"
import os
import sys
import tempfile
import unittest

import yaml
from kapitan.cached import reset_cache
from kapitan.cli import main
from kapitan.inputs.helm import Helm

helm_binding_exists = True
try:
    from kapitan.inputs.helm.helm_binding import ffi  # this statement will raise ImportError if binding not available
except ImportError:
    helm_binding_exists = False


@unittest.skipUnless(helm_binding_exists, "helm binding is not available")
class HelmInputTest(unittest.TestCase):
    def setUp(self):
        os.chdir(os.path.join("tests", "test_resources"))

    def test_render_chart(self):
        temp_dir = tempfile.mkdtemp()
        chart_path = "charts/acs-engine-autoscaler"
        helm = Helm(None, None, None)
        error_message = helm.render_chart(chart_path, temp_dir)
        self.assertFalse(error_message)
        self.assertTrue(os.path.isfile(os.path.join(temp_dir, "acs-engine-autoscaler", "templates", "secrets.yaml")))
        self.assertTrue(os.path.isfile(os.path.join(temp_dir, "acs-engine-autoscaler", "templates", "deployment.yaml")))

    def test_error_invalid_char_dir(self):
        chart_path = "non-existent"
        temp_dir = tempfile.mkdtemp()
        helm = Helm(None, None, None)
        error_message = helm.render_chart(chart_path, temp_dir)
        self.assertTrue("no such file or directory" in error_message)

    def test_compile_chart(self):
        temp = tempfile.mkdtemp()
        sys.argv = ["kapitan", "compile", "--output-path", temp, "-t", "acs-engine-autoscaler"]
        main()
        self.assertTrue(os.path.isfile(
            os.path.join(temp, "compiled", "acs-engine-autoscaler", "acs-engine-autoscaler", "templates", "secrets.yaml")))

    def test_compile_subcharts(self):
        temp = tempfile.mkdtemp()
        sys.argv = ["kapitan", "compile", "--output-path", temp, "-t", "istio"]
        main()
        self.assertTrue(os.path.isdir(os.path.join(temp, "compiled", "istio", "istio", "charts")))
        self.assertTrue(os.path.isdir(os.path.join(temp, "compiled", "istio", "istio", "templates")))

    def test_compile_multiple_targets(self):
        temp = tempfile.mkdtemp()
        sys.argv = ["kapitan", "compile", "--output-path", temp,
                    "-t", "acs-engine-autoscaler", "nginx-ingress", "-p", "2"]
        main()
        self.assertTrue(os.path.isfile(
            os.path.join(temp, "compiled", "acs-engine-autoscaler", "acs-engine-autoscaler",
                         "templates", "secrets.yaml")))
        self.assertTrue(os.path.isfile(
            os.path.join(temp, "compiled", "nginx-ingress", "nginx-ingress", "templates", "clusterrolebinding.yaml")))

    def test_compile_multiple_charts_per_target(self):
        temp = tempfile.mkdtemp()
        sys.argv = ["kapitan", "compile", "--output-path", temp,
                    "-t", "nginx-istio"]
        main()
        self.assertTrue(os.path.isdir(os.path.join(temp, "compiled", "nginx-istio", "istio", "templates")))
        self.assertTrue(os.path.isdir(os.path.join(temp, "compiled", "nginx-istio", "nginx-ingress", "templates")))

    def test_compile_with_helm_values(self):
        temp = tempfile.mkdtemp()
        sys.argv = ["kapitan", "compile", "--output-path", temp, "-t", "nginx-ingress"]
        main()
        controller_deployment_file = \
            os.path.join(temp, "compiled", "nginx-ingress", "nginx-ingress", "templates", "controller-deployment.yaml")
        self.assertTrue(os.path.isfile(controller_deployment_file))
        with open(controller_deployment_file, 'r') as fp:
            manifest = yaml.safe_load(fp.read())
            name = manifest['metadata']['name']
            self.assertEqual(name, '-nginx-ingress-my-controller')

    def test_compile_with_helm_params(self):
        temp = tempfile.mkdtemp()
        sys.argv = ["kapitan", "compile", "--output-path", temp, "-t", "nginx-ingress-helm-params"]
        with open('inventory/targets/nginx-ingress-helm-params.yml', 'r') as fp:
            manifest = yaml.safe_load(fp.read())
            helm_params = manifest['parameters']['kapitan']['compile'][0]['helm_params']
            release_name = helm_params['release_name']
            namespace = helm_params['namespace']

        main()
        controller_deployment_file = os.path.join(temp, "compiled", "nginx-ingress-helm-params",
                                                  "nginx-ingress", "templates", "controller-deployment.yaml")

        self.assertTrue(os.path.isfile(controller_deployment_file))
        with open(controller_deployment_file, 'r') as fp:
            manifest = yaml.safe_load(fp.read())
            container = manifest['spec']['template']['spec']['containers'][0]
            property = container['args'][4]
            self.assertEqual(property, '--configmap={}/{}'.format(namespace,
                                                                  release_name + '-nginx-ingress-my-controller'))

    def tearDown(self):
        os.chdir('../../')
        reset_cache()



#!/usr/bin/env python3
#
# Copyright 2019 The Kapitan Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging
import os
import json

from kapitan.inputs.base import InputType, CompiledFile
from kapitan.resources import search_imports, resource_callbacks
from kapitan.utils import jsonnet_file, prune_empty

logger = logging.getLogger(__name__)


class Jsonnet(InputType):
    def __init__(self, compile_path, search_paths, ref_controller):
        super().__init__("jsonnet", compile_path, search_paths, ref_controller)

    def compile_file(self, file_path, compile_path, ext_vars, **kwargs):
        """
        Write file_path (jsonnet evaluated) items as files to compile_path.
        ext_vars will be passed as parameters to jsonnet_file()
        kwargs:
            output: default 'yaml', accepts 'json'
            prune: default False, accepts True
            reveal: default False, set to reveal refs on compile
            target_name: default None, set to current target being compiled
            indent: default 2
        """
        def _search_imports(cwd, imp):
            return search_imports(cwd, imp, self.search_paths)

        json_output = jsonnet_file(file_path, import_callback=_search_imports,
                                   native_callbacks=resource_callbacks(self.search_paths),
                                   ext_vars=ext_vars)
        output_obj = json.loads(json_output)

        output = kwargs.get('output', 'yaml')
        prune = kwargs.get('prune', False)
        reveal = kwargs.get('reveal', False)
        target_name = kwargs.get('target_name', None)
        indent = kwargs.get('indent', 2)

        if prune:
            output_obj = prune_empty(output_obj)
            logger.debug("Pruned output for: %s", file_path)

        for item_key, item_value in output_obj.items():
            # write each item to disk
            if output == 'json':
                file_path = os.path.join(compile_path, '%s.%s' % (item_key, output))
                with CompiledFile(file_path, self.ref_controller, mode="w", reveal=reveal, target_name=target_name,
                                  indent=indent) as fp:
                    fp.write_json(item_value)
            elif output == 'yaml':
                file_path = os.path.join(compile_path, '%s.%s' % (item_key, "yml"))
                with CompiledFile(file_path, self.ref_controller, mode="w", reveal=reveal, target_name=target_name,
                                  indent=indent) as fp:
                    fp.write_yaml(item_value)
            elif output == 'plain':
                file_path = os.path.join(compile_path, '%s' % item_key)
                with CompiledFile(file_path, self.ref_controller, mode="w", reveal=reveal, target_name=target_name,
                                  indent=indent) as fp:
                    fp.write(item_value)
            else:
                raise ValueError('output is neither "json", "yaml" or "plain"')

    def default_output_type(self):
        return "yaml"



#!/usr/bin/env python3
#
# Copyright 2019 The Kapitan Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import os
import sys
import unittest
import tempfile

from kapitan.cached import reset_cache
from kapitan.cli import main
from kapitan.dependency_manager.base import fetch_git_source, fetch_http_source, fetch_git_dependency, DEPENDENCY_OUTPUT_CONFIG


class DependencyManagerTest(unittest.TestCase):
    def setUp(self):
        os.chdir(os.path.join(os.getcwd(), "tests", "test_resources"))

    def test_fetch_http_sources(self):
        temp_dir = tempfile.mkdtemp()
        http_sources = [
            "https://raw.githubusercontent.com/deepmind/kapitan/master/examples/docker/components/jsonnet/jsonnet.jsonnet",
            "https://raw.githubusercontent.com/deepmind/kapitan/master/examples/docker/components/kadet/__init__.py"
        ]

        for source in http_sources:
            fetch_http_source(source, temp_dir)

        self.assertTrue(os.path.isfile(os.path.join(temp_dir, '1c3a08e6' + 'jsonnet.jsonnet')))
        self.assertTrue(os.path.isfile(os.path.join(temp_dir, 'aff45ec8' + '__init__.py')))

    def test_fetch_git_sources(self):
        temp_dir = tempfile.mkdtemp()
        # TODO: also test git ssh urls
        git_source = 'https://github.com/deepmind/kapitan.git'
        fetch_git_source(git_source, temp_dir)
        self.assertTrue(os.path.isdir(os.path.join(temp_dir, 'kapitan.git', 'kapitan')))

    def test_clone_repo_subdir(self):
        temp_dir = tempfile.mkdtemp()
        output_dir = tempfile.mkdtemp()
        source = 'https://github.com/deepmind/kapitan.git'
        dep = [
                {
                    "output_path": os.path.join(output_dir, "subdir"),
                    "ref": "master",
                    "subdir": "tests"
                }
        ]
        fetch_git_dependency((source, dep), temp_dir)
        self.assertTrue(os.path.isdir(os.path.join(output_dir, "subdir")))

    def test_compile_fetch(self):
        temp = tempfile.mkdtemp()
        DEPENDENCY_OUTPUT_CONFIG["root_dir"] = temp
        sys.argv = ["kapitan", "compile", "--fetch", "--output-path", temp, "-t",
                    "nginx", "nginx-dev", "-p", "4"]
        main()
        self.assertTrue(os.path.isdir(os.path.join(temp, "components", "tests")))
        self.assertTrue(os.path.isdir(os.path.join(temp, "components", "acs-engine-autoscaler")))
        self.assertTrue(os.path.isdir(os.path.join(temp, "components", "kapitan-repository")))
        self.assertTrue(os.path.isdir(os.path.join(temp, "components", "source")))

    def tearDown(self):
        os.chdir('../../')
        reset_cache()



#!/usr/bin/env python3
#
# Copyright 2019 The Kapitan Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"kapitan resources"

import base64
import errno
import gzip
import io
import json
import jsonschema
import logging
import os
from functools import partial

import kapitan.cached as cached
import yaml
from kapitan import __file__ as kapitan_install_path
from kapitan.errors import CompileError, InventoryError
from kapitan.utils import render_jinja2_file, sha256_string

import reclass
import reclass.core
from reclass.errors import NotFoundError, ReclassException

logger = logging.getLogger(__name__)

try:
    from yaml import CSafeLoader as YamlLoader
except ImportError:
    from yaml import SafeLoader as YamlLoader


def resource_callbacks(search_paths):
    """
    Returns a dict with all the functions to be used
    on the native_callbacks keyword on jsonnet
    search_paths can be used by the native functions to access files
    """

    return {"jinja2_render_file": (("name", "ctx"),
                                   partial(jinja2_render_file, search_paths)),
            "inventory": (("target", "inv_path"),
                          partial(inventory, search_paths)),
            "file_read": (("name",),
                          partial(read_file, search_paths)),
            "sha256_string": (("obj",), sha256_string),
            "gzip_b64": (("obj",), gzip_b64),
            "yaml_dump": (("obj",), yaml_dump),
            "yaml_load": (("name",),
                          partial(yaml_load, search_paths)),
            "jsonschema_validate": (("obj", "schema_obj"), jsonschema_validate),
            }


def jsonschema_validate(obj, schema_obj):
    """
    validates obj with a schema_obj jsonschema
    returns an object with keys:
    - valid (true/false)
    - reason (will have jsonschema validation error when not valid)
    """
    _obj = json.loads(obj)
    _schema_obj = json.loads(schema_obj)
    try:
        jsonschema.validate(_obj, _schema_obj, format_checker=jsonschema.FormatChecker())
        return {"valid": True, "reason": ""}
    except jsonschema.ValidationError as e:
        return {"valid": False, "reason": ""+str(e)}


def yaml_dump(obj):
    """Dumps jsonnet obj as yaml"""
    _obj = json.loads(obj)
    return yaml.safe_dump(_obj, default_flow_style=False)


def gzip_b64(obj):
    """returns base64-encoded gzip-compressed obj"""
    obj_bytes = obj.encode("UTF-8")
    buf = io.BytesIO()
    with gzip.GzipFile(fileobj=buf, mode='wb', compresslevel=9, mtime=0) as f:
        f.write(obj_bytes)
    compressed_obj = buf.getvalue()
    return base64.b64encode(compressed_obj).decode("UTF-8")


def jinja2_render_file(search_paths, name, ctx):
    """
    Render jinja2 file name with context ctx.
    search_paths is used to find the file name
    as there is a limitation with jsonnet's native_callback approach:
    one can't access the current directory being evaluated
    """
    ctx = json.loads(ctx)
    _full_path = ""

    for path in search_paths:
        _full_path = os.path.join(path, name)
        logger.debug("jinja2_render_file trying file %s", _full_path)
        if os.path.exists(_full_path):
            logger.debug("jinja2_render_file found file at %s", _full_path)
            try:
                return render_jinja2_file(_full_path, ctx)
            except Exception as e:
                raise CompileError("Jsonnet jinja2 failed to render {}: {}".format(_full_path, e))

    raise IOError("jinja2 failed to render, could not find file: {}".format(_full_path))


def yaml_load(search_paths, name):
    """returns content of yaml file as json string"""
    for path in search_paths:
        _full_path = os.path.join(path, name)
        logger.debug("yaml_load trying file %s", _full_path)
        if os.path.exists(_full_path) and (name.endswith(".yml") or name.endswith(".yaml")):
            logger.debug("yaml_load found file at %s", _full_path)
            try:
                with open(_full_path) as f:
                    return json.dumps(yaml.safe_load(f.read()))
            except Exception as e:
                raise CompileError("Parse yaml failed to parse {}: {}".format(_full_path, e))

    raise IOError("could not find any input yaml file: {}".format(_full_path))


def read_file(search_paths, name):
    """return content of file in name"""
    for path in search_paths:
        full_path = os.path.join(path, name)
        logger.debug("read_file trying file %s", full_path)
        if os.path.exists(full_path):
            logger.debug("read_file found file at %s", full_path)
            with io.open(full_path, newline='') as f:
                return f.read()

    raise IOError("Could not find file {}".format(name))


def search_imports(cwd, import_str, search_paths):
    """
    This is only to be used as a callback for the jsonnet API!
    - cwd is the import context $CWD,
    - import_str is the "bla.jsonnet" in 'import "bla.jsonnet"'
    - search_paths is the location where to look for import_str if not in cwd
    The only supported parameters are cwd and import_str, so search_paths
    needs to be closured.
    """
    basename = os.path.basename(import_str)
    full_import_path = os.path.join(cwd, import_str)

    if not os.path.exists(full_import_path):
        # if import_str not found, search in install_path
        install_path = os.path.dirname(kapitan_install_path)
        _full_import_path = os.path.join(install_path, import_str)
        # if found, set as full_import_path
        if os.path.exists(_full_import_path):
            full_import_path = _full_import_path
            logger.debug("import_str: %s found in search_path: %s",
                         import_str, install_path)
        else:
            # if import_str not found, search in search_paths
            for path in search_paths:
                _full_import_path = os.path.join(path, import_str)
                # if found, set as full_import_path
                if os.path.exists(_full_import_path):
                    full_import_path = _full_import_path
                    logger.debug("import_str: %s found in search_path: %s",
                                 import_str, path)
                    break

    # if the above search did not find anything, let jsonnet error
    # with a non existent import
    normalised_path = os.path.normpath(full_import_path)

    logger.debug("cwd:%s import_str:%s basename:%s -> norm:%s",
                 cwd, import_str, basename, normalised_path)

    normalised_path_content = ""
    with open(normalised_path) as f:
        normalised_path_content = f.read()

    return normalised_path, normalised_path_content


def inventory(search_paths, target, inventory_path="inventory/"):
    """
    Reads inventory (set by inventory_path) in search_paths.
    set nodes_uri to change reclass nodes_uri the default value
    set target to None to return all target in the inventory
    Returns a dictionary with the inventory for target
    """

    full_inv_path = ""
    inv_path_exists = False
    for path in search_paths:
        full_inv_path = os.path.join(path, inventory_path)
        if os.path.exists(full_inv_path):
            inv_path_exists = True
            break

    if not inv_path_exists:
        raise InventoryError("Inventory not found in search paths: {}".format(search_paths))

    if target is None:
        return inventory_reclass(full_inv_path)["nodes"]

    return inventory_reclass(full_inv_path)["nodes"][target]


def inventory_reclass(inventory_path):
    """
    Runs a reclass inventory in inventory_path
    (same output as running ./reclass.py -b inv_base_uri/ --inventory)
    Will attempt to read reclass config from 'reclass-config.yml' otherwise
    it will failback to the default config.
    Returns a reclass style dictionary
    """

    if not cached.inv:
        reclass_config = {'storage_type': 'yaml_fs',
                          'inventory_base_uri': inventory_path,
                          'nodes_uri': os.path.join(inventory_path, 'targets'),
                          'classes_uri': os.path.join(inventory_path, 'classes'),
                          'compose_node_name': False
                          }

        try:
            cfg_file = os.path.join(inventory_path, 'reclass-config.yml')
            with open(cfg_file) as reclass_cfg:
                reclass_config = yaml.load(reclass_cfg, Loader=YamlLoader)
                # normalise relative nodes_uri and classes_uri paths
                for uri in ('nodes_uri', 'classes_uri'):
                    uri_val = reclass_config.get(uri)
                    uri_path = os.path.join(inventory_path, uri_val)
                    normalised_path = os.path.normpath(uri_path)
                    reclass_config.update({uri: normalised_path})
                logger.debug("Using reclass inventory config at: %s", cfg_file)
        except IOError as ex:
            # If file does not exist, ignore
            if ex.errno == errno.ENOENT:
                logger.debug("Using reclass inventory config defaults")

        try:
            storage = reclass.get_storage(reclass_config['storage_type'],
                                          reclass_config['nodes_uri'],
                                          reclass_config['classes_uri'],
                                          reclass_config['compose_node_name'])
            class_mappings = reclass_config.get('class_mappings')  # this defaults to None (disabled)
            _reclass = reclass.core.Core(storage, class_mappings, reclass.settings.Settings(reclass_config))

            cached.inv = _reclass.inventory()
        except ReclassException as e:
            if isinstance(e, NotFoundError):
                logger.error("Inventory reclass error: inventory not found")
            else:
                logger.error("Inventory reclass error: %s", e.message)
            raise InventoryError(e.message)

    return cached.inv



#!/usr/bin/env python3
#
# Copyright 2019 The Kapitan Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"init tests"

import logging
import os
import tempfile
import unittest

from kapitan.initialiser import initialise_skeleton

logging.basicConfig(level=logging.CRITICAL, format="%(message)s")
logger = logging.getLogger(__name__)


class InitTest(unittest.TestCase):
    def test_init(self):
        with tempfile.TemporaryDirectory() as tmp_dir:
            initialise_skeleton(directory=tmp_dir)

            template_dir = os.path.join(
                os.getcwd(), 'kapitan', 'inputs', 'templates')

            diff_files = []

            for root, dirs, files in os.walk(tmp_dir):
                diff_files += files

            for root, dirs, files in os.walk(template_dir):
                for f in files:
                    if f in diff_files:
                        diff_files.remove(f)

            self.assertEqual(len(diff_files), 0)



# Copyright 2019 The Kapitan Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"references module"

import base64
import errno
import hashlib
import json
import logging
import os
import re
import sys
from functools import lru_cache

import yaml
from kapitan.errors import (RefBackendError, RefError, RefFromFuncError,
                            RefHashMismatchError)
from kapitan.refs.functions import eval_func
from kapitan.utils import PrettyDumper, list_all_paths

try:
    from yaml import CSafeLoader as YamlLoader
except ImportError:
    from yaml import SafeLoader as YamlLoader

logger = logging.getLogger(__name__)

# e.g. ?{ref:my/secret/token} or ?{ref:my/secret/token||func:param1:param2}
REF_TOKEN_TAG_PATTERN = r"(\?{([\w\:\.\-\/@]+)([\|\|\w\:\.\-\/]+)?=*})"
REF_TOKEN_SUBVAR_PATTERN = r"(@[\w\.\-\_]+)"


class PlainRef(object):
    def __init__(self, data, **kwargs):
        """
        writes plain data
        """
        self.type_name = 'plain'
        self.encoding = kwargs.get('encoding', 'original')
        self.data = data

    def reveal(self):
        return self.data

    def compile(self):
        return self.data

    def __str__(self):
        return "{}".format(self.__dict__)

    @classmethod
    def from_path(cls, ref_full_path, **kwargs):
        """
        return a new PlainRef from file at ref_full_path
        """
        try:
            with open(ref_full_path) as fp:
                obj = yaml.load(fp, Loader=YamlLoader)
                _kwargs = {key: value for key, value in obj.items() if key not in ('data',)}
                kwargs.update(_kwargs)
                return cls(obj['data'], **kwargs)

        except IOError as ex:
            if ex.errno == errno.ENOENT:
                return None

    @classmethod
    def from_params(cls, data, ref_params):
        """
        Return new PlainRef from data and ref_params
        """
        # default encoding to 'original'
        # TODO encoding needs a better place other than kwargs
        encoding = ref_params.kwargs.get('encoding', 'original')
        if encoding == 'original':
            return cls(data.encode(), **ref_params.kwargs)
        elif encoding == 'base64':
            # data already bytes encoded
            return cls(data, **ref_params.kwargs)

    def dump(self):
        """
        Returns dict with keys/values to be serialised.
        """
        return {"data": self.data.decode(), "encoding": self.encoding,
                "type": self.type_name}


class RefParams(object):
    def __init__(self, *args, **kwargs):
        "pack params for new Refs from functions"
        self.args = args
        self.kwargs = kwargs


class PlainRefBackend(object):
    def __init__(self, path, ref_type=PlainRef):
        "Get and create PlainRefs"
        self.path = path
        self.type_name = 'plain'
        self.ref_type = ref_type

    def __getitem__(self, ref_path):
        # remove the substring notation, if any
        ref_file_path = re.sub(REF_TOKEN_SUBVAR_PATTERN, "", ref_path)
        full_ref_path = os.path.join(self.path, ref_file_path)
        ref = self.ref_type.from_path(full_ref_path)

        if ref is not None:
            ref.path = ref_path
            ref_path_data = "{}{}".format(ref_file_path, ref.data)
            ref.hash = hashlib.sha256(ref_path_data.encode()).hexdigest()
            ref.token = "{}:{}:{}".format(ref.type_name, ref.path, ref.hash[:8])
            return ref

        raise KeyError(ref_path)

    def __setitem__(self, ref_path, ref_obj):
        assert(isinstance(ref_obj, self.ref_type))
        full_ref_path = os.path.join(self.path, ref_path)
        os.makedirs(os.path.dirname(full_ref_path), exist_ok=True)

        with open(full_ref_path, 'w') as fp:
            yaml.safe_dump(ref_obj.dump(), stream=fp, default_flow_style=False)

    def __contains__(self, ref_path):
        try:
            self.__getitem__(ref_path)
            return True
        except KeyError:
            return False

    def __iter__(self):
        for full_path in list_all_paths(self.path):
            ref_path = full_path[len(self.path):]
            yield ref_path

    def iteritems(self):
        for k in self.__iter__():
            try:
                v = self.__getitem__(k)
                yield (k, v)
            except KeyError:
                pass


class Revealer(object):
    def __init__(self, ref_controller):
        "reveal files and objects"
        self.ref_controller = ref_controller
        self.regex = re.compile(REF_TOKEN_TAG_PATTERN)

    def reveal_path(self, path):
        "detects if path is file or dir, returns list of RevealObj"
        if os.path.isfile(path):
            content, content_type = self._reveal_file(path)
            return [RevealedObj(content, content_type)]
        elif os.path.isdir(path):
            content_yaml, content_json, content_raw = self._reveal_dir(path)
            revealed_objs = []
            if content_yaml != '':
                revealed_objs.append(RevealedObj(content_yaml, 'yaml'))
            elif content_json != '':
                revealed_objs.append(RevealedObj(content_json, 'json'))
            elif content_raw != '':
                revealed_objs.append(RevealedObj(content_raw, 'raw'))
            return revealed_objs
        else:
            raise FileNotFoundError(path)

    def _reveal_file(self, filename):
        """
        detects filename by extension (.yml/.yaml, .json or otherwise raw)
        reveals secrets in content
        """
        if filename.endswith('.yml') or filename.endswith('.yaml'):
            logger.debug("Revealer: revealing yml file: %s", filename)
            with open(filename) as fp:
                obj = [o for o in yaml.load_all(fp, Loader=YamlLoader)]
                rev_obj = self.reveal_obj(obj)
                return yaml.dump_all(rev_obj, Dumper=PrettyDumper, default_flow_style=False,
                                     explicit_start=True), 'yaml'
        elif filename.endswith('.json'):
            logger.debug("Revealer: revealing json file: %s", filename)
            with open(filename) as fp:
                obj = json.load(fp)
                rev_obj = self.reveal_obj(obj)
                return json.dumps(rev_obj, indent=4, sort_keys=True), 'json'
        else:
            logger.debug("Revealer: revealing raw file: %s", filename)
            return self.reveal_raw_file(filename), 'raw'

    def _reveal_dir(self, dirname):
        """
        returns tuple with yaml, json, and raw concatenated output for revealed file types
        it does not walk through the dir structure as it skips directories inside dirname
        """
        out_yaml = ''
        out_json = ''
        out_raw = ''
        # find yaml/json/raw files and concatenate output per type
        for f in os.listdir(dirname):
            full_path = os.path.join(dirname, f)
            if not os.path.isfile(full_path):
                pass
            if f.endswith('.yml') or f.endswith('.yaml'):
                out, _ = self._reveal_file(full_path)
                out_yaml += out
            elif f.endswith('.json'):
                out, _ = self._reveal_file(full_path)
                out_json += self._reveal_file(full_path)
            else:
                out, _ = self._reveal_file(full_path)
                out_raw += out

        return out_yaml, out_json, out_raw

    def _reveal_replace_match(self, match_obj):
        """returns decrypted value from tag in match_obj"""
        tag, _, _ = match_obj.groups()
        m = re.search(REF_TOKEN_SUBVAR_PATTERN, tag)
        if m is None:
            # no subvar pattern is found
            logger.debug("Revealer: no sub-variable path was matched in \"{}\"".format(tag))
            return self._reveal_tag_without_subvar(tag)

        else:
            subvar_path = m.groups()
            # strip away the @
            subvar_path = subvar_path[0][1:]
            logger.debug("Revealer: sub-variable path \"{}\" matched in tag {}".
                         format(subvar_path, tag))
            tag_without_yaml_path = re.sub(REF_TOKEN_SUBVAR_PATTERN, "", tag)

            plaintext = self._reveal_tag_without_subvar(tag_without_yaml_path)
            revealed_yaml = yaml.load(plaintext, Loader=YamlLoader)
            if not isinstance(revealed_yaml, dict):
                raise RefError("revealed secret is not in yaml, cannot access sub-variable"
                               " at {}".format(subvar_path))
            return self._get_value_in_yaml_path(revealed_yaml, subvar_path)

    @lru_cache(maxsize=256)
    def _reveal_tag_without_subvar(self, tag_without_subvar):
        ref = self.ref_controller[tag_without_subvar]
        logger.debug("Revealer: revealing tag {} for the first time".format(tag_without_subvar))
        return ref.reveal()

    def _get_value_in_yaml_path(self, d, yaml_path):
        """using the sub-variable path as nested keys, returns the value in the dictionary"""
        keys = yaml_path.split('.')
        value = d
        for key in keys:
            value = value[key]

        return value

    def _compile_replace_match_with_args(self, **kwargs):
        """returns compile_replace_match function with kwargs"""
        def compile_replace_match(match_obj):
            """returns compile ref value from tag in match_obj"""
            tag, _, _ = match_obj.groups()
            try:
                ref = self.ref_controller[tag]
                return ref.compile()
            # if refs needs to be created from func:
            except RefFromFuncError:
                # create ref from func with kwargs RefParams
                self.ref_controller[tag] = RefParams(**kwargs)
                ref = self.ref_controller[tag]
                return ref.compile()
            # if refs don't exist:
            except KeyError:
                raise RefError("Could not find ref: {}".format(tag))

        return compile_replace_match

    def reveal_raw_file(self, filename):
        """
        read filename and reveal content (per line search and replace) with refs
        set filename=None to read stdin
        returns string with revealed content
        """
        out_raw = ''
        if filename is None:
            for line in sys.stdin:
                revealed = self.regex.sub(self._reveal_replace_match, line)
                out_raw += revealed
        else:
            with open(filename) as fp:
                for line in fp:
                    revealed = self.regex.sub(self._reveal_replace_match, line)
                    out_raw += revealed
        return out_raw

    def compile_raw(self, data, **kwargs):
        """
        read data and compile refs (per line search and replace) with output of ref .compile() method
        kwargs are passed to ref compile function
        returns string with compile content
        """
        compiled = self.regex.sub(self._compile_replace_match_with_args(**kwargs), data)
        return compiled

    def reveal_raw(self, data):
        """
        read data and reveal content (per line search and replace) with refs
        returns string with revealed content
        """
        revealed = self.regex.sub(self._reveal_replace_match, data)
        return revealed

    def reveal_obj(self, obj):
        """recursively updates obj with revealed refs"""
        if isinstance(obj, dict):
            for k, v in obj.items():
                obj[k] = self.reveal_obj(v)
        elif isinstance(obj, list):
            obj = [self.reveal_obj(item) for item in obj]
        elif isinstance(obj, str):
            obj = self.regex.sub(self._reveal_replace_match, obj)
        return obj

    def compile_obj(self, obj, **kwargs):
        """
        recursively updates obj with compiled refs
        kwargs are passed to ref compile function
        """
        if isinstance(obj, dict):
            for k, v in obj.items():
                obj[k] = self.compile_obj(v, **kwargs)
        elif isinstance(obj, list):
            obj = [self.compile_obj(item, **kwargs) for item in obj]
        elif isinstance(obj, str):
            obj = self.regex.sub(self._compile_replace_match_with_args(**kwargs), obj)
        return obj


class RevealedObj(object):
    "Content and content type object type"
    def __init__(self, content, content_type):
        self.content = content
        self.content_type = content_type


class RefController(object):
    def __init__(self, path):
        """
        gets and sets tags for ref type objects.
        auto registers backends
        """
        self.backends = {}
        self.path = path

    def register_backend(self, backend):
        "register backend type"
        assert(isinstance(backend, PlainRefBackend))
        self.backends[backend.type_name] = backend

    def _get_backend(self, type_name):
        "imports and registers backend according to type_name"
        try:
            return self.backends[type_name]
        except KeyError:
            if type_name == 'plain':
                from kapitan.refs.base import PlainRefBackend
                self.register_backend(PlainRefBackend(self.path))
            elif type_name == 'base64':
                from kapitan.refs.base64 import Base64RefBackend
                self.register_backend(Base64RefBackend(self.path))
            elif type_name == 'gpg':
                from kapitan.refs.secrets.gpg import GPGBackend
                self.register_backend(GPGBackend(self.path))
            elif type_name == 'gkms':
                from kapitan.refs.secrets.gkms import GoogleKMSBackend
                self.register_backend(GoogleKMSBackend(self.path))
            elif type_name == 'awskms':
                from kapitan.refs.secrets.awskms import AWSKMSBackend
                self.register_backend(AWSKMSBackend(self.path))
            elif type_name == 'vaultkv':
                from kapitan.refs.secrets.vaultkv import VaultBackend
                self.register_backend(VaultBackend(self.path))
            else:
                raise RefBackendError('no backend for ref type: {}'.format(type_name))
        return self.backends[type_name]

    def tag_type(self, tag):
        "returns ref type for tag"
        # ?{ref:my/secret/token} or ?{ref:my/secret/token||func:param1:param2} or ?{ref:my/secret/token:deadbeef}
        _, token, _ = self.tag_params(tag)
        return self.token_type(token)

    def tag_params(self, tag):
        "returns tag parameter tuple with (tag, token, func_str)"
        match = re.match(REF_TOKEN_TAG_PATTERN, tag)
        if match:
            tag, token, func_str = match.groups()
            return tag, token, func_str
        else:
            raise RefError("{}: is not a valid tag".format(tag))

    def token_type(self, token):
        "returns ref type for token"
        attrs = token.split(':')
        type_name = attrs[0]

        # force type_name to register
        self._get_backend(type_name)

        return self.backends[type_name].ref_type

    def token_type_name(self, token):
        "returns ref type name for token"
        attrs = token.split(':')
        type_name = attrs[0]

        return type_name

    def _get_from_token(self, token):
        attrs = token.split(':')

        # "type_name:path/to/ref"
        if len(attrs) == 2:
            type_name = attrs[0]
            path = attrs[1]
            backend = self._get_backend(type_name)
            return backend[path]

        # "type_name:path/to/ref:n0c0ffee"
        elif len(attrs) == 3:
            type_name = attrs[0]
            path = attrs[1]
            hash = attrs[2]
            backend = self._get_backend(type_name)
            ref = backend[path]
            if ref.hash[:8] == hash:
                return ref
            else:
                raise RefHashMismatchError("{}: hash does not match with: {}".format(token, ref.token))
        else:
            return None

    def _set_to_token(self, token, ref_obj):
        attrs = token.split(':')

        if len(attrs) == 2:
            type_name = attrs[0]
            path = attrs[1]
            backend = self._get_backend(type_name)
            assert(isinstance(ref_obj, backend.ref_type))
            backend[path] = ref_obj
        else:
            raise RefError("{}: is not a valid token".format(token))

    def _eval_func_str(self, ctx, func_str):
        """
        evals and updates context ctx for func_str
        returns evaluated ctx
        """
        assert(func_str.startswith('||'))
        funcs = func_str[2:].split('|')

        for func in funcs:
            func_name, *func_params = func.strip().split(':')
            if func_name == 'base64':  # not a real function
                ctx.encode_base64 = True
            else:
                try:
                    eval_func(func_name, ctx, *func_params)
                except KeyError:
                    raise RefError("{}: unknown function".format(func_name))

        return ctx

    def __getitem__(self, key):
        # ?{ref:my/secret/token} or ?{ref:my/secret/token|func:param1:param2} or ?{ref:my/secret/token:deadbeef}
        tag, token, func_str = self.tag_params(key)

        # if there is no function, grab token
        if func_str is None:
            ref = self._get_from_token(token)
            if ref:
                return ref
        else:
            if re.search(REF_TOKEN_SUBVAR_PATTERN, token) is not None:
                raise RefError("Ref: secrets with sub-variables must be created manually")
            # if there is a function, try grabbing token
            try:
                ref = self._get_from_token(token)
                if ref:
                    return ref
            except KeyError:
                # if func is set and ref doesnt exist,
                # raise RefFromFuncError to indicate new ref needs to be created
                raise RefFromFuncError('{}: does not exist and must be '
                                       'created from function: {}'.format(token, func_str))

        raise KeyError("{}: ref not found".format(tag))

    def __setitem__(self, key, value):
        # ?{ref:my/secret/token} or ?{ref:my/secret/token|func:param1:param2}
        tag, token, func_str = self.tag_params(key)

        if func_str is None and isinstance(value, self.token_type(token)):
            return self._set_to_token(token, value)
        # if function is set, ensure value is RefParams instance
        elif func_str is not None and isinstance(value, RefParams):
            # run _eval_func_str, create ref_obj and run _set_to_token()
            ctx = FunctionContext(None)
            ctx.encode_base64 = False
            ctx.ref_controller = self
            ctx.token = token

            self._eval_func_str(ctx, func_str)
            ref_type = self.token_type(token)

            if ctx.encode_base64:
                b64_data = base64.b64encode(ctx.data.encode())
                # TODO encoding needs a better place other than kwargs
                value.kwargs['encoding'] = 'base64'
                ref_obj = ref_type.from_params(b64_data, value)
            else:
                ref_obj = ref_type.from_params(ctx.data, value)

            return self._set_to_token(token, ref_obj)


class FunctionContext(object):
    def __init__(self, data):
        "Carry context across function evaluation"
        self.data = data



# Copyright 2019 The Kapitan Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"reference functions"

import base64
import hashlib
import logging
import secrets  # python secrets module

from cryptography.hazmat.backends import default_backend
from cryptography.hazmat.primitives import serialization
from cryptography.hazmat.primitives.asymmetric import rsa
from kapitan.errors import RefError

logger = logging.getLogger(__name__)


def eval_func(func_name, ctx, *func_params):

    func_lookup = {
        'randomstr': randomstr,
        'sha256': sha256,
        'rsa': rsa_private_key,
        'rsapublic': rsa_public_key,
        'reveal': reveal
    }

    return func_lookup[func_name](ctx, *func_params)


def randomstr(ctx, nbytes=''):
    """
    generates a URL-safe text string, containing nbytes random bytes
    sets it to ctx.data
    """
    if nbytes:
        nbytes = int(nbytes)
        # Generate twice the amount of bytes asked for
        # and then trim the string to nbytes length if it's longer
        ctx.data = secrets.token_urlsafe(2 * nbytes)[:nbytes]

    else:
        ctx.data = secrets.token_urlsafe()


def sha256(ctx, salt=''):
    """sets ctx.data to salted sha256 hexdigest for input_value"""
    if ctx.data:
        salted_input_value = salt + ":" + ctx.data
        ctx.data = hashlib.sha256(salted_input_value.encode()).hexdigest()
    else:
        raise RefError("Ref error: eval_func: nothing to sha256 hash; try "
                       "something like '|randomstr|sha256'")


def rsa_private_key(ctx, key_size='4096'):
    """sets ctx.data to a RSA private key of key_size, default 4096"""
    rsa_key_size = int(key_size)

    key = rsa.generate_private_key(public_exponent=65537, key_size=rsa_key_size,
                                   backend=default_backend())

    ctx.data = str(key.private_bytes(
        encoding=serialization.Encoding.PEM,
        format=serialization.PrivateFormat.PKCS8,
        encryption_algorithm=serialization.NoEncryption()
    ), "utf-8")


def rsa_public_key(ctx):
    """Derives RSA public key from revealed private key"""
    if not ctx.data:
        raise RefError("Ref error: eval_func: RSA public key cannot be derived; try "
                       "something like '|reveal:path/to/encrypted_private_key|rsapublic'")

    data_dec = ctx.data
    if ctx.ref_encoding == 'base64':
        data_dec = base64.b64decode(data_dec).decode()

    private_key = serialization.load_pem_private_key(data_dec.encode(), password=None, backend=default_backend())
    public_key = private_key.public_key()

    ctx.data = str(public_key.public_bytes(
                   encoding=serialization.Encoding.PEM,
                   format=serialization.PublicFormat.SubjectPublicKeyInfo
                   ), "UTF-8")


def reveal(ctx, secret_path):
    """
    decrypt and return data from secret_path
    """
    token_type_name = ctx.ref_controller.token_type_name(ctx.token)
    secret_tag = "?{{{}:{}}}".format(token_type_name, secret_path)
    try:
        ref_obj = ctx.ref_controller[secret_tag]
        ctx.ref_encoding = ref_obj.encoding
        ctx.data = ref_obj.reveal()
    except KeyError:
        raise RefError("|reveal function error: {secret_path} file in {}|reveal:{secret_path} does not exist"
                       .format(ctx.token, secret_path=secret_path))



#!/usr/bin/env python3
#
# Copyright 2019 The Kapitan Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"linter tests"

import logging
import unittest

from kapitan.lint import start_lint

logging.basicConfig(level=logging.CRITICAL, format="%(message)s")
logger = logging.getLogger(__name__)


class LinterTest(unittest.TestCase):
    def test_lint(self):
        num_issues_found = start_lint(fail_on_warning=False,
                            skip_class_checks=False,
                            skip_yamllint=False,
                            inventory_path="./tests/test_resources/inventory",
                            search_secrets=True,
                            secrets_path="./tests/test_resources/secrets",
                            compiled_path="./tests/test_resources/compiled")
        desired_output = 3
        self.assertEqual(num_issues_found, desired_output)



# Copyright 2019 The Kapitan Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"kapitan error classes"


class KapitanError(Exception):
    """generic kapitan error"""
    pass


class CompileError(KapitanError):
    """compile error"""
    pass


class InventoryError(KapitanError):
    """inventory error"""
    pass


class SecretError(KapitanError):
    """secrets error"""
    pass


class RefError(KapitanError):
    """ref error"""
    pass


class RefBackendError(KapitanError):
    """ref backend error"""
    pass


class RefFromFuncError(KapitanError):
    """ref from func error"""
    pass


class RefHashMismatchError(KapitanError):
    """ref has mismatch error"""
    pass


class HelmBindingUnavailableError(KapitanError):
    """helm input is used when the binding is not available"""
    pass


class HelmTemplateError(KapitanError):
    pass


class GitSubdirNotFoundError(KapitanError):
    """git dependency subdir not found error"""
    pass


class RequestUnsuccessfulError(KapitanError):
    """request error"""
    pass


class KubernetesManifestValidationError(KapitanError):
    """kubernetes manifest schema validation error"""
    pass



#!/usr/bin/env python3
#
# Copyright 2019 The Kapitan Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"gpg secrets tests"

import os
import tempfile
import unittest

import kapitan.cached as cached
from cryptography.hazmat.backends import default_backend
from cryptography.hazmat.primitives import serialization
from kapitan.refs.base import RefController, RefParams, Revealer
from kapitan.refs.secrets.gpg import (GPG_KWARGS, GPG_TARGET_FINGERPRINTS,
                                      GPGSecret, gpg_obj)

# set GNUPGHOME for test_cli
GNUPGHOME = tempfile.mkdtemp()
os.environ["GNUPGHOME"] = GNUPGHOME

gpg_obj(gnupghome=GNUPGHOME)

KEY = cached.gpg_obj.gen_key(cached.gpg_obj.gen_key_input(key_type="RSA",
                                                          key_length=2048,
                                                          passphrase="testphrase"))
KEY2 = cached.gpg_obj.gen_key(cached.gpg_obj.gen_key_input(key_type="RSA",
                                                           key_length=2048,
                                                           passphrase="testphrase"))
GPG_TARGET_FINGERPRINTS["KEY"] = KEY.fingerprint
GPG_KWARGS["passphrase"] = "testphrase"

REFS_HOME = tempfile.mkdtemp()
REF_CONTROLLER = RefController(REFS_HOME)
REVEALER = Revealer(REF_CONTROLLER)


class GPGSecretsTest(unittest.TestCase):
    "Test GPG secrets"

    def test_gpg_write_reveal(self):
        "write secret, confirm secret file exists, reveal and compare content"
        tag = '?{gpg:secret/sauce}'
        REF_CONTROLLER[tag] = GPGSecret("super secret value", [{'fingerprint': KEY.fingerprint}])
        self.assertTrue(os.path.isfile(os.path.join(REFS_HOME, 'secret/sauce')))

        file_with_secret_tags = tempfile.mktemp()
        with open(file_with_secret_tags, 'w') as fp:
            fp.write('I am a file with a ?{gpg:secret/sauce}')
        revealed = REVEALER.reveal_raw_file(file_with_secret_tags)
        self.assertEqual("I am a file with a super secret value", revealed)

    def test_gpg_base64_write_reveal(self):
        """
        write secret for base64 encoded content, confirm secret file exists,
        reveal and compare content
        """
        tag = '?{gpg:secret/sauce2}'
        REF_CONTROLLER[tag] = GPGSecret("super secret value", [{'fingerprint': KEY.fingerprint}],
                                        encode_base64=True)
        self.assertTrue(os.path.isfile(os.path.join(REFS_HOME, 'secret/sauce2')))

        file_with_secret_tags = tempfile.mktemp()
        with open(file_with_secret_tags, 'w') as fp:
            fp.write('I am a file with a ?{gpg:secret/sauce2}')
        revealed = REVEALER.reveal_raw_file(file_with_secret_tags)
        self.assertEqual("I am a file with a c3VwZXIgc2VjcmV0IHZhbHVl", revealed)

    def test_gpg_function_rsa(self):
        "write rsa (private and public), confirm secret file exists, reveal and check"

        tag = '?{gpg:secret/rsa||rsa}'
        REF_CONTROLLER[tag] = RefParams()
        self.assertTrue(os.path.isfile(os.path.join(REFS_HOME, 'secret/rsa')))

        file_with_secret_tags = tempfile.mktemp()
        with open(file_with_secret_tags, 'w') as fp:
            fp.write('?{gpg:secret/rsa}')
        revealed = REVEALER.reveal_raw_file(file_with_secret_tags)
        try:
            serialization.load_pem_private_key(revealed.encode(), password=None, backend=default_backend())
        except ValueError:
            raise Exception("Failed to decode RSA private key")

        REVEALER._reveal_tag_without_subvar.cache_clear()
        # Test with parameter key_size=2048
        tag = '?{gpg:secret/rsa||rsa:2048}'
        REF_CONTROLLER[tag] = RefParams()
        revealed = REVEALER.reveal_raw_file(file_with_secret_tags)

        try:
            private_key = serialization.load_pem_private_key(revealed.encode(), password=None,
                                                             backend=default_backend())
        except ValueError:
            raise Exception("Failed to decode RSA private key")

        self.assertEqual(private_key.key_size, 2048)

        # Test rsapublic with previous private key as the parameter
        tag_rsapublic = '?{gpg:secret/rsapublic||reveal:secret/rsa|rsapublic}'
        REF_CONTROLLER[tag_rsapublic] = RefParams()
        self.assertTrue(os.path.isfile(os.path.join(REFS_HOME, 'secret/rsa')))

        file_with_secret_tags = tempfile.mktemp()
        with open(file_with_secret_tags, 'w') as fp:
            fp.write('?{gpg:secret/rsapublic}')
        revealed = REVEALER.reveal_raw_file(file_with_secret_tags)
        self.assertEqual(revealed.splitlines()[0], "-----BEGIN PUBLIC KEY-----")

    def test_gpg_update_recipients(self):
        """
        update existing base64'd secret with another recipient, confirm content is the same
        """
        tag = '?{gpg:secret/sauce_with_fingerprints}'
        secret = GPGSecret("super secret value", [{'fingerprint': KEY.fingerprint}], encode_base64=True)
        REF_CONTROLLER[tag] = secret
        ref = REF_CONTROLLER[tag]
        self.assertTrue(os.path.isfile(os.path.join(REFS_HOME, 'secret/sauce_with_fingerprints')))
        self.assertTrue(len(ref.recipients), 1)

        file_with_secret_tags = tempfile.mktemp()
        with open(file_with_secret_tags, 'w') as fp:
            fp.write('I am a file with a ?{gpg:secret/sauce_with_fingerprints}')

        new_recipients = [{'fingerprint': KEY.fingerprint}, {'fingerprint': KEY2.fingerprint}]
        ref.update_recipients(new_recipients)
        ref = REF_CONTROLLER[tag]

        self.assertTrue(len(ref.recipients), 2)
        revealed = REVEALER.reveal_raw_file(file_with_secret_tags)
        self.assertEqual("I am a file with a c3VwZXIgc2VjcmV0IHZhbHVl", revealed)



#!/usr/bin/env python3
#
# Copyright 2019 The Kapitan Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"initialiser module"

import logging
import os
from distutils.dir_util import copy_tree

logger = logging.getLogger(__name__)


def initialise_skeleton(directory):
    """ Initialises a directory with a recommended skeleton structure
    Args:
        directory (string): path which to initialise, directory is assumed to exist
    """

    current_pwd = os.path.dirname(__file__)
    templates_directory = os.path.join(current_pwd, 'inputs', 'templates')

    copy_tree(templates_directory, directory)

    logger.info("Populated {} with:".format(directory))
    for dirName, subdirList, fileList in os.walk(directory):
        logger.info('{}'.format(dirName))
        for fname in fileList:
            logger.info('\t {}'.format(fname))
        # Remove the first entry in the list of sub-directories
        # if there are any sub-directories present
        if len(subdirList) > 0:
            del subdirList[0]



from cffi import FFI


def main():
    ffi = FFI()

    # declare functions to export
    ffi.cdef("""
        char* renderChart(char* p0, char* p1, char* p2, char* p3, char* p4, char* p5);
        void free(void* ptr);
    """)

    ffi.set_source(
        "helm_binding",  # specify name for importing this module
        None
    )

    ffi.compile(verbose=True)


if __name__ == '__main__':
    main()



#!/usr/bin/env python3
#
# Copyright 2019 The Kapitan Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"jsonnet tests"

import json
import jsonschema
import unittest
import os

from kapitan.resources import yaml_dump, gzip_b64, yaml_load, jsonschema_validate
from kapitan.utils import sha256_string, prune_empty


class JsonnetNativeFuncsTest(unittest.TestCase):
    def test_yaml_dump(self):
        """dump json string to yaml"""
        yaml = yaml_dump("{\"key\":\"value\"}")
        self.assertEqual(yaml, "key: value\n")

    def test_yaml_load(self):
        """
            This tests the yaml_load function.
            It converts the yaml file in test_resources/ to a json string
        """
        current_pwd = os.path.dirname(__file__)
        json = yaml_load([current_pwd], "test_resources/test_yaml_load.yaml")
        expected_output = """{"test": {"key": "value", "array": ["ele1", "ele2"]}}"""
        self.assertEqual(json, expected_output)

    def test_sha256_string(self):
        """sha256 hex digest for string"""
        hash = sha256_string("test")
        self.assertEqual(hash, "9f86d081884c7d659a2feaa0c55ad015a3bf4f1b2b0b822cd15d6c15b0f00a08")

    def test_gzip_b64(self):
        """base64-encoded gzip-compression for string"""
        gzip = gzip_b64("test")
        self.assertEqual(gzip, "H4sIAAAAAAAC/ytJLS4BAAx+f9gEAAAA")

    def test_prune_empty(self):
        """Remove empty lists and empty dictionaries from dict"""
        dictionary = {"hello": "world", "array": [1, 2], "foo": {}, "bar": []}
        pruned = prune_empty(dictionary)
        self.assertEqual(pruned, {"hello": "world", "array": [1, 2]})

    def test_jsonschema_valid(self):
        """validate valid obj with jsonschema"""
        dictionary = {"msg": "hello, world!", "array": [1, 2]}
        schema = {
                    "type": "object",
                    "properties": {
                        "msg": {"type": "string"},
                        "array": {"type": "array", "contains": {"type": "number"}},
                        }
                    }
        validation = jsonschema_validate(json.dumps(dictionary), json.dumps(schema))

        self.assertTrue(validation["valid"])
        self.assertEqual(validation["reason"], "")

    def test_jsonschema_invalid(self):
        """validate invalid obj with jsonschema"""
        dictionary = {"msg": "hello, world!", "array": ["a", "b", "c"]}
        schema = {
                    "type": "object",
                    "properties": {
                        "msg": {"type": "string"},
                        "array": {"type": "array", "contains": {"type": "number"}},
                        }
                    }
        validation = jsonschema_validate(json.dumps(dictionary), json.dumps(schema))

        self.assertFalse(validation["valid"])
        self.assertNotEqual(validation["reason"], "")



# Copyright 2019 The Kapitan Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"base64 ref module"

import base64
import errno
import logging
import yaml

from kapitan.refs.base import PlainRef, PlainRefBackend

try:
    from yaml import CSafeLoader as YamlLoader
except ImportError:
    from yaml import SafeLoader as YamlLoader

logger = logging.getLogger(__name__)


class Base64Ref(PlainRef):
    def __init__(self, data, from_base64=False, **kwargs):
        """
        writes data
        set from_base64 to load already base64 encoded data
        """
        super().__init__(data, **kwargs)
        self.type_name = 'base64'
        self.encoding = kwargs.get('encoding', 'original')
        # TODO data should be bytes only
        if from_base64:
            self.data = data
        else:
            self.data = base64.b64encode(data).decode()

    def reveal(self):
        # TODO data should be bytes only
        return base64.b64decode(self.data).decode()

    def compile(self):
        # XXX will only work if object read via backend
        return "?{{{}:{}:{}}}".format(self.type_name, self.path, self.hash[:8])

    @classmethod
    def from_path(cls, ref_full_path, **kwargs):
        """
        return a new Base64Ref from file at ref_full_path
        the data key in the file must be base64 encoded
        """
        try:
            with open(ref_full_path) as fp:
                obj = yaml.load(fp, Loader=YamlLoader)
                _kwargs = {key: value for key, value in obj.items() if key not in ('data', 'from_base64')}
                kwargs.update(_kwargs)
                return cls(obj['data'], from_base64=True, **kwargs)

        except IOError as ex:
            if ex.errno == errno.ENOENT:
                return None

    def dump(self):
        """
        Returns dict with keys/values to be serialised.
        """
        return {"data": self.data, "encoding": self.encoding,
                "type": self.type_name}


class Base64RefBackend(PlainRefBackend):
    def __init__(self, path, ref_type=Base64Ref):
        super().__init__(path, ref_type)
        self.type_name = 'base64'



#!/usr/bin/env python3
#
# Copyright 2019 The Kapitan Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"version check tests"

import unittest

from kapitan.utils import compare_versions


class VersionTest(unittest.TestCase):
    "Test version checks"

    def test_version_equal(self):
        dot_kapitan_version = "0.22.0"
        current_version = "0.22.0"
        result = compare_versions(dot_kapitan_version, current_version)
        self.assertEqual(result, "equal")

    def test_version_rc_equal(self):
        dot_kapitan_version = "0.22.0-rc.1"
        current_version = "0.22.0-rc.1"
        result = compare_versions(dot_kapitan_version, current_version)
        self.assertEqual(result, "equal")

    def test_version_major_equal(self):
        dot_kapitan_version = "0.22"
        current_version = "0.22.1"
        result = compare_versions(dot_kapitan_version, current_version)
        self.assertEqual(result, "equal")

    def test_version_one_rc_major_equal(self):
        dot_kapitan_version = "0.22"
        current_version = "0.22.1-rc.1"
        result = compare_versions(dot_kapitan_version, current_version)
        self.assertEqual(result, "equal")

    def test_version_greater(self):
        dot_kapitan_version = "0.22.1"
        current_version = "0.22.0"
        result = compare_versions(dot_kapitan_version, current_version)
        self.assertEqual(result, "greater")

    def test_version_one_rc_greater(self):
        dot_kapitan_version = "0.22.0"
        current_version = "0.22.0-rc.1"
        result = compare_versions(dot_kapitan_version, current_version)
        self.assertEqual(result, "greater")

    def test_version_both_rc_greater(self):
        dot_kapitan_version = "0.22.1-rc.1"
        current_version = "0.22.0-rc.1"
        result = compare_versions(dot_kapitan_version, current_version)
        self.assertEqual(result, "greater")

    def test_version_both_rc_major_greater(self):
        dot_kapitan_version = "0.23.0-rc.1"
        current_version = "0.22.0-rc.1"
        result = compare_versions(dot_kapitan_version, current_version)
        self.assertEqual(result, "greater")

    def test_version_lower(self):
        dot_kapitan_version = "0.22.0"
        current_version = "0.22.1"
        result = compare_versions(dot_kapitan_version, current_version)
        self.assertEqual(result, "lower")

    def test_version_one_rc_lower(self):
        dot_kapitan_version = "0.22.0-rc.1"
        current_version = "0.22.0"
        result = compare_versions(dot_kapitan_version, current_version)
        self.assertEqual(result, "lower")

    def test_version_both_rc_lower(self):
        dot_kapitan_version = "0.22.0-rc.1"
        current_version = "0.22.1-rc.1"
        result = compare_versions(dot_kapitan_version, current_version)
        self.assertEqual(result, "lower")

    def test_version_both_rc_major_lower(self):
        dot_kapitan_version = "0.22.0-rc.1"
        current_version = "0.23.0-rc.1"
        result = compare_versions(dot_kapitan_version, current_version)
        self.assertEqual(result, "lower")



#!/usr/bin/env python3
#
# Copyright 2019 The Kapitan Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import os
from shutil import copyfile
import sys
import unittest
import tempfile
import yaml
from kapitan.cached import reset_cache
from kapitan.cli import main
from kapitan.errors import KubernetesManifestValidationError
from kapitan.validator.kubernetes_validator import KubernetesManifestValidator, FILE_PATH_FORMAT


class KubernetesValidatorTest(unittest.TestCase):
    def setUp(self):
        os.chdir(os.path.join('examples', 'kubernetes'))
        self.cache_dir = tempfile.mkdtemp()
        self.validator = KubernetesManifestValidator(self.cache_dir)

    def test_download_and_cache(self):
        kind = 'service'
        version = '1.14.0'
        downloaded_schema = self.validator._get_schema_from_web(kind, version)
        self.validator._cache_schema(kind, version, downloaded_schema)
        self.assertTrue(os.path.isfile(os.path.join(self.cache_dir, FILE_PATH_FORMAT.format(version, kind))))

    def test_load_from_cache(self):
        kind = 'deployment'
        version = '1.11.0'  # different version just for testing
        downloaded_schema = self.validator._get_schema_from_web(kind, version)
        self.validator._cache_schema(kind, version, downloaded_schema)
        self.assertIsInstance(self.validator._get_cached_schema(kind, version), dict)

    def test_validate(self):
        service_manifest_string = """
            apiVersion: v1
            kind: Service
            metadata:
              name: my-service
            spec:
              selector:
                app: MyApp
              ports:
              - protocol: TCP
                port: 80
                targetPort: 9376
        """
        manifest_path = os.path.join(self.cache_dir, 'service_manifest.yaml')
        with open(manifest_path, 'w') as fp:
            fp.write(service_manifest_string)
        self.validator.validate([manifest_path], kind='service', version='1.14.0')

        with self.assertRaises(KubernetesManifestValidationError):
            self.validator.validate([manifest_path], kind='deployment', version='1.14.0')

    def test_validate_command_pass(self):
        sys.argv = ['kapitan', 'validate', '--schemas-path', self.cache_dir]
        try:
            main()
        except SystemExit:
            self.fail("Kubernetes manifest validation error raised unexpectedly")

    def test_validate_command_fail(self):
        file_name_format = 'inventory/classes/component/mysql{}.yml'
        original_file = file_name_format.format('')
        copied_file = file_name_format.format('_copy')
        copyfile(original_file, copied_file)
        wrong_manifest_kind = 'deployment'
        with open(original_file, 'r') as fp:
            d = yaml.safe_load(fp)
            # change kind from service to deployment
            d['parameters']['kapitan']['validate'][0]['kind'] = wrong_manifest_kind
        with open(original_file, 'w') as fp:
            yaml.dump(d, fp, default_flow_style=False)

        sys.argv = ['kapitan', 'validate', '--schemas-path', self.cache_dir]
        with self.assertRaises(SystemExit), self.assertLogs(logger='kapitan.targets', level='ERROR') as log:
            try:
                main()
            finally:
                # copy back the original file
                copyfile(copied_file, original_file)
                os.remove(copied_file)
        self.assertTrue(' '.join(log.output).find("invalid '{}' manifest".format(wrong_manifest_kind)) != -1)

    def test_validate_after_compile(self):
        sys.argv = ['kapitan', 'compile', '-t', 'minikube-mysql', '--validate', '--schemas-path', self.cache_dir]
        try:
            main()
        except SystemExit:
            self.fail("Kubernetes manifest validation error raised unexpectedly")

    def tearDown(self):
        os.chdir('../../')
        reset_cache()



#!/usr/bin/env python3
#
# Copyright 2019 The Kapitan Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"inventory tests"

import unittest
from kapitan.resources import inventory


class InventoryTargetTest(unittest.TestCase):
    def test_inventory_target(self):
        inv = inventory(["examples/kubernetes"], "minikube-es")
        self.assertEqual(inv["parameters"]["cluster"]["name"], "minikube")

    def test_inventory_all_targets(self):
        inv = inventory(["examples/kubernetes"], None)
        self.assertNotEqual(inv.get("minikube-es"), None)
